from __future__ import print_function

import cStringIO
import os
import sys
import traceback

import py

from rpython.flowspace.model import (FunctionGraph, Constant, Variable)
from rpython.rlib import rstackovf
from rpython.rlib.objectmodel import (ComputedIntSymbolic, CDefinedIntSymbolic,
    Symbolic)
# intmask is used in an exec'd code block
from rpython.rlib.rarithmetic import (ovfcheck, is_valid_int, intmask,
    r_uint, r_longlong, r_ulonglong, r_longlonglong)
from rpython.rtyper.lltypesystem import lltype, llmemory, lloperation, llheap
from rpython.rtyper import rclass
from rpython.tool.ansi_print import AnsiLogger

# by default this logger's output is disabled.
# e.g. tests can then switch on logging to get more help
# for failing tests
log = AnsiLogger('llinterp')
log.output_disabled = True


class LLException(Exception):

    # .error_value is used only by tests: in particular,
    # test_exceptiontransform uses it to check what is the return value of the
    # function in case of exception
    UNDEFINED_ERROR_VALUE = object() # sentinel for self.error_value

    def __init__(self, *args, **kwargs):
        "NOT_RPYTHON"
        Exception.__init__(self, *args)
        self.error_value = kwargs.pop('error_value', self.UNDEFINED_ERROR_VALUE)
        if kwargs:
            raise TypeError('unexpected keyword arguments: %s' % kwargs.keys())

    def __str__(self):
        etype = self.args[0]
        #evalue = self.args[1]
        if len(self.args) > 2:
            f = cStringIO.StringIO()
            original_type, original_value, original_tb = self.args[2]
            traceback.print_exception(original_type, original_value, original_tb,
                                      file=f)
            extra = '\n' + f.getvalue().rstrip('\n')
            extra = extra.replace('\n', '\n | ') + '\n `------'
        else:
            extra = ''
        return '<LLException %r%s>' % (type_name(etype), extra)

class LLFatalError(Exception):
    def __str__(self):
        return ': '.join([str(x) for x in self.args])

class LLAssertFailure(Exception):
    pass


def type_name(etype):
    return ''.join(etype.name.chars)

class LLInterpreter(object):
    """ low level interpreter working with concrete values. """

    current_interpreter = None

    def __init__(self, typer, tracing=True, exc_data_ptr=None):
        self.bindings = {}
        self.typer = typer
        # 'heap' is module or object that provides malloc, etc for lltype ops
        self.heap = llheap
        self.exc_data_ptr = exc_data_ptr
        self.frame_stack = []
        self.tracer = None
        self.frame_class = LLFrame
        if tracing:
            self.tracer = Tracer()

    def eval_graph(self, graph, args=(), recursive=False):
        llframe = self.frame_class(graph, args, self)
        if self.tracer and not recursive:
            global tracer1
            tracer1 = self.tracer
            self.tracer.start()
        retval = None
        self.traceback_frames = []
        old_frame_stack = self.frame_stack[:]
        prev_interpreter = LLInterpreter.current_interpreter
        LLInterpreter.current_interpreter = self
        try:
            try:
                retval = llframe.eval()
            except LLException as e:
                log.error("LLEXCEPTION: %s" % (e, ))
                self.print_traceback()
                if self.tracer:
                    self.tracer.dump('LLException: %s\n' % (e,))
                raise
            except Exception as e:
                if getattr(e, '_go_through_llinterp_uncaught_', False):
                    raise
                log.error("AN ERROR OCCURED: %s" % (e, ))
                self.print_traceback()
                if self.tracer:
                    line = str(e)
                    if line:
                        line = ': ' + line
                    line = '* %s' % (e.__class__.__name__,) + line
                    self.tracer.dump(line + '\n')
                raise
        finally:
            LLInterpreter.current_interpreter = prev_interpreter
            assert old_frame_stack == self.frame_stack
            if self.tracer:
                if retval is not None:
                    self.tracer.dump('   ---> %r\n' % (retval,))
                if not recursive:
                    self.tracer.stop()
        return retval

    def print_traceback(self):
        frames = self.traceback_frames
        frames.reverse()
        self.traceback_frames = []
        lines = []
        for frame in frames:
            logline = frame.graph.name + "()"
            if frame.curr_block is None:
                logline += " <not running yet>"
                lines.append(logline)
                continue
            try:
                logline += " " + self.typer.annotator.annotated[frame.curr_block].func.__module__
            except (KeyError, AttributeError, TypeError):
                logline += " <unknown module>"
            lines.append(logline)
            for i, operation in enumerate(frame.curr_block.operations):
                if i == frame.curr_operation_index:
                    logline = "E  %s"
                else:
                    logline = "   %s"
                lines.append(logline % (operation, ))
        if self.tracer:
            self.tracer.dump('Traceback\n', bold=True)
            for line in lines:
                self.tracer.dump(line + '\n')
        for line in lines:
            log.traceback(line)

    def get_tlobj(self):
        try:
            return self._tlobj
        except AttributeError:
            from rpython.rtyper.lltypesystem import rffi
            PERRNO = rffi.CArrayPtr(rffi.INT)
            fake_p_errno = lltype.malloc(PERRNO.TO, 1, flavor='raw', zero=True,
                                         track_allocation=False)
            self._tlobj = {'RPY_TLOFS_p_errno': fake_p_errno,
                           #'thread_ident': ...,
                           }
            return self._tlobj

    def find_roots(self, is_minor=False):
        """Return a list of the addresses of the roots."""
        #log.findroots("starting")
        roots = []
        for frame in reversed(self.frame_stack):
            #log.findroots("graph", frame.graph.name)
            frame.find_roots(roots)
            # If a call is done with 'is_minor=True', we can stop after the
            # first frame in the stack that was already seen by the previous
            # call with 'is_minor=True'.  (We still need to trace that frame,
            # but not its callers.)
            if is_minor:
                if getattr(frame, '_find_roots_already_seen', False):
                    break
                frame._find_roots_already_seen = True
        return roots

    def find_exception(self, exc):
        assert isinstance(exc, LLException)
        klass, inst = exc.args[0], exc.args[1]
        for cls in enumerate_exceptions_top_down():
            if "".join(klass.name.chars) == cls.__name__:
                return cls
        raise ValueError("couldn't match exception, maybe it"
                      " has RPython attributes like OSError?")

    def get_transformed_exc_data(self, graph):
        if hasattr(graph, 'exceptiontransformed'):
            return graph.exceptiontransformed
        if getattr(graph, 'rgenop', False):
            return self.exc_data_ptr
        return None

    def _store_exception(self, exc):
        raise PleaseOverwriteStoreException("You just invoked ll2ctypes callback without overwriting _store_exception on llinterpreter")

class PleaseOverwriteStoreException(Exception):
    pass

def checkptr(ptr):
    assert isinstance(lltype.typeOf(ptr), lltype.Ptr)

def checkadr(addr):
    assert lltype.typeOf(addr) is llmemory.Address


class LLFrame(object):
    def __init__(self, graph, args, llinterpreter):
        assert not graph or isinstance(graph, FunctionGraph)
        self.graph = graph
        self.args = args
        self.llinterpreter = llinterpreter
        self.heap = llinterpreter.heap
        self.bindings = {}
        self.curr_block = None
        self.curr_operation_index = 0
        self.alloca_objects = []

    def newsubframe(self, graph, args):
        return self.__class__(graph, args, self.llinterpreter)

    # _______________________________________________________
    # variable setters/getters helpers

    def clear(self):
        self.bindings.clear()

    def fillvars(self, block, values):
        vars = block.inputargs
        assert len(vars) == len(values), (
                   "block %s received %d args, expected %d" % (
                    block, len(values), len(vars)))
        for var, val in zip(vars, values):
            self.setvar(var, val)

    def setvar(self, var, val):
        if var.concretetype is not lltype.Void:
            try:
                val = lltype.enforce(var.concretetype, val)
            except TypeError:
                assert False, "type error: input value of type:\n\n\t%r\n\n===> variable of type:\n\n\t%r\n" % (lltype.typeOf(val), var.concretetype)
        assert isinstance(var, Variable)
        self.bindings[var] = val

    def setifvar(self, var, val):
        if isinstance(var, Variable):
            self.setvar(var, val)

    def getval(self, varorconst):
        try:
            val = varorconst.value
        except AttributeError:
            val = self.bindings[varorconst]
        if isinstance(val, ComputedIntSymbolic):
            val = val.compute_fn()
        if varorconst.concretetype is not lltype.Void:
            try:
                val = lltype.enforce(varorconst.concretetype, val)
            except TypeError:
                assert False, "type error: %r val from %r var/const" % (lltype.typeOf(val), varorconst.concretetype)
        return val

    # _______________________________________________________
    # other helpers
    def getoperationhandler(self, opname):
        ophandler = getattr(self, 'op_' + opname, None)
        if ophandler is None:
            # try to import the operation from opimpl.py
            ophandler = lloperation.LL_OPERATIONS[opname].fold
            setattr(self.__class__, 'op_' + opname, staticmethod(ophandler))
        return ophandler
    # _______________________________________________________
    # evaling functions

    def eval(self):
        graph = self.graph
        tracer = self.llinterpreter.tracer
        if tracer:
            tracer.enter(graph)
        self.llinterpreter.frame_stack.append(self)
        try:
            try:
                nextblock = graph.startblock
                args = self.args
                while 1:
                    self.clear()
                    self.fillvars(nextblock, args)
                    nextblock, args = self.eval_block(nextblock)
                    if nextblock is None:
                        for obj in self.alloca_objects:
                            obj._obj._free()
                        return args
            except Exception:
                self.llinterpreter.traceback_frames.append(self)
                raise
        finally:
            leavingframe = self.llinterpreter.frame_stack.pop()
            assert leavingframe is self
            if tracer:
                tracer.leave()

    def eval_block(self, block):
        """ return (nextblock, values) tuple. If nextblock
            is None, values is the concrete return value.
        """
        self.curr_block = block
        e = None

        try:
            for i, op in enumerate(block.operations):
                self.curr_operation_index = i
                self.eval_operation(op)
        except LLException as e:
            if op is not block.raising_op:
                raise
        except RuntimeError as e:
            rstackovf.check_stack_overflow()
            # xxx fish fish fish for proper etype and evalue to use
            rtyper = self.llinterpreter.typer
            bk = rtyper.annotator.bookkeeper
            classdef = bk.getuniqueclassdef(rstackovf._StackOverflow)
            exdata = rtyper.exceptiondata
            evalue = exdata.get_standard_ll_exc_instance(rtyper, classdef)
            etype = exdata.fn_type_of_exc_inst(evalue)
            e = LLException(etype, evalue)
            if op is not block.raising_op:
                raise e

        # determine nextblock and/or return value
        if len(block.exits) == 0:
            # return block
            tracer = self.llinterpreter.tracer
            if len(block.inputargs) == 2:
                # exception
                if tracer:
                    tracer.dump('raise')
                etypevar, evaluevar = block.getvariables()
                etype = self.getval(etypevar)
                evalue = self.getval(evaluevar)
                # watch out, these are _ptr's
                raise LLException(etype, evalue)
            resultvar, = block.getvariables()
            result = self.getval(resultvar)
            exc_data = self.llinterpreter.get_transformed_exc_data(self.graph)
            if exc_data:
                # re-raise the exception set by this graph, if any
                etype = exc_data.exc_type
                if etype:
                    evalue = exc_data.exc_value
                    if tracer:
                        tracer.dump('raise')
                    exc_data.exc_type = lltype.typeOf(etype)._defl()
                    exc_data.exc_value = lltype.typeOf(evalue)._defl()
                    raise LLException(etype, evalue, error_value=result)
            if tracer:
                tracer.dump('return')
            return None, result
        elif block.exitswitch is None:
            # single-exit block
            assert len(block.exits) == 1
            link = block.exits[0]
        elif block.canraise:
            link = block.exits[0]
            if e:
                exdata = self.llinterpreter.typer.exceptiondata
                cls = e.args[0]
                inst = e.args[1]
                for link in block.exits[1:]:
                    assert issubclass(link.exitcase, py.builtin.BaseException)
                    if self.op_direct_call(exdata.fn_exception_match,
                                           cls, link.llexitcase):
                        self.setifvar(link.last_exception, cls)
                        self.setifvar(link.last_exc_value, inst)
                        break
                else:
                    # no handler found, pass on
                    raise e
        else:
            llexitvalue = self.getval(block.exitswitch)
            if block.exits[-1].exitcase == "default":
                defaultexit = block.exits[-1]
                nondefaultexits = block.exits[:-1]
                assert defaultexit.llexitcase is None
            else:
                defaultexit = None
                nondefaultexits = block.exits
            for link in nondefaultexits:
                if link.llexitcase == llexitvalue:
                    break   # found -- the result is in 'link'
            else:
                if defaultexit is None:
                    raise ValueError("exit case %r not found in the exit links "
                                     "of %r" % (llexitvalue, block))
                else:
                    link = defaultexit
        return link.target, [self.getval(x) for x in link.args]

    def eval_operation(self, operation):
        tracer = self.llinterpreter.tracer
        if tracer:
            tracer.dump(str(operation))
        ophandler = self.getoperationhandler(operation.opname)
        # XXX slighly unnice but an important safety check
        if operation.opname == 'direct_call':
            assert isinstance(operation.args[0], Constant)
        elif operation.opname == 'indirect_call':
            assert isinstance(operation.args[0], Variable)
        if getattr(ophandler, 'specialform', False):
            retval = ophandler(*operation.args)
        else:
            vals = [self.getval(x) for x in operation.args]
            if getattr(ophandler, 'need_result_type', False):
                vals.insert(0, operation.result.concretetype)
            try:
                retval = ophandler(*vals)
            except LLException as e:
                # safety check check that the operation is allowed to raise that
                # exception
                if operation.opname in lloperation.LL_OPERATIONS:
                    canraise = lloperation.LL_OPERATIONS[operation.opname].canraise
                    if Exception not in canraise:
                        exc = self.llinterpreter.find_exception(e)
                        for canraiseexc in canraise:
                            if issubclass(exc, canraiseexc):
                                break
                        else:
                            raise TypeError("the operation %s is not expected to raise %s" % (operation, exc))

                # for exception-transformed graphs, store the LLException
                # into the exc_data used by this graph
                exc_data = self.llinterpreter.get_transformed_exc_data(
                    self.graph)
                if exc_data:
                    etype = e.args[0]
                    evalue = e.args[1]
                    exc_data.exc_type = etype
                    exc_data.exc_value = evalue
                    retval = e.error_value
                    if retval is LLException.UNDEFINED_ERROR_VALUE:
                        from rpython.translator import exceptiontransform
                        # if we are here it means that the exception was
                        # caused by a builtin op such as int_add_ovf (i.e.,
                        # NOT a call): in this case, we just use the default
                        # error_value
                        T = operation.result.concretetype
                        retval = exceptiontransform.default_error_value(T)
                else:
                    raise
        self.setvar(operation.result, retval)
        if tracer:
            if retval is None:
                tracer.dump('\n')
            else:
                tracer.dump('   ---> %r\n' % (retval,))

    def make_llexception(self, exc=None):
        if exc is None:
            original = sys.exc_info()
            exc = original[1]
            # it makes no sense to convert some exception classes that
            # just mean something buggy crashed
            if isinstance(exc, (AssertionError, AttributeError,
                                TypeError, NameError,
                                KeyboardInterrupt, SystemExit,
                                ImportError, SyntaxError)):
                raise original[0], original[1], original[2]     # re-raise it
            # for testing the JIT (see ContinueRunningNormally) we need
            # to let some exceptions introduced by the JIT go through
            # the llinterpreter uncaught
            if getattr(exc, '_go_through_llinterp_uncaught_', False):
                raise original[0], original[1], original[2]     # re-raise it
            extraargs = (original,)
        else:
            extraargs = ()
        typer = self.llinterpreter.typer
        exdata = typer.exceptiondata
        evalue = exdata.get_standard_ll_exc_instance_by_class(exc.__class__)
        etype = self.op_direct_call(exdata.fn_type_of_exc_inst, evalue)
        raise LLException(etype, evalue, *extraargs)

    def invoke_callable_with_pyexceptions(self, fptr, *args):
        obj = fptr._obj
        try:
            return obj._callable(*args)
        except LLException as e:
            raise
        except Exception as e:
            if getattr(e, '_go_through_llinterp_uncaught_', False):
                raise
            if getattr(obj, '_debugexc', False):
                log.ERROR('The llinterpreter got an '
                          'unexpected exception when calling')
                log.ERROR('the external function %r:' % (fptr,))
                log.ERROR('%s: %s' % (e.__class__.__name__, e))
                if self.llinterpreter.tracer:
                    self.llinterpreter.tracer.flush()
                import sys
                from rpython.translator.tool.pdbplus import PdbPlusShow
                PdbPlusShow(None).post_mortem(sys.exc_info()[2])
            self.make_llexception()

    def find_roots(self, roots):
        #log.findroots(self.curr_block.inputargs)
        vars = []
        for v in self.curr_block.inputargs:
            if isinstance(v, Variable):
                vars.append(v)
        for op in self.curr_block.operations[:self.curr_operation_index]:
            vars.append(op.result)

        for v in vars:
            TYPE = v.concretetype
            if isinstance(TYPE, lltype.Ptr) and TYPE.TO._gckind == 'gc':
                roots.append(_address_of_local_var(self, v))

    # __________________________________________________________
    # misc LL operation implementations

    def op_debug_view(self, *ll_objects):
        from rpython.translator.tool.lltracker import track
        track(*ll_objects)

    def op_debug_assert(self, x, msg):
        if not x:
            raise LLAssertFailure(msg)

    def op_debug_assert_not_none(self, x):
        if not x:
            raise LLAssertFailure("ll_assert_not_none() failed")

    def op_debug_fatalerror(self, ll_msg, ll_exc=None):
        msg = ''.join(ll_msg.chars)
        if ll_exc is None:
            raise LLFatalError(msg)
        else:
            ll_exc_type = lltype.cast_pointer(rclass.OBJECTPTR, ll_exc).typeptr
            raise LLFatalError(msg, LLException(ll_exc_type, ll_exc))

    def op_debug_llinterpcall(self, pythonfunction, *args_ll):
        try:
            return pythonfunction(*args_ll)
        except:
            self.make_llexception()

    def op_debug_forked(self, *args):
        raise NotImplementedError

    def op_debug_start_traceback(self, *args):
        pass    # xxx write debugging code here?

    def op_debug_reraise_traceback(self, *args):
        pass    # xxx write debugging code here?

    def op_debug_record_traceback(self, *args):
        pass    # xxx write debugging code here?

    def op_debug_print_traceback(self, *args):
        pass    # xxx write debugging code here?

    def op_debug_catch_exception(self, *args):
        pass    # xxx write debugging code here?

    def op_jit_marker(self, *args):
        pass

    def op_jit_record_exact_class(self, *args):
        pass

    def op_jit_record_exact_value(self, *args):
        pass

    def op_jit_conditional_call(self, *args):
        raise NotImplementedError("should not be called while not jitted")

    def op_jit_conditional_call_value(self, *args):
        raise NotImplementedError("should not be called while not jitted")

    def op_get_exception_addr(self, *args):
        raise NotImplementedError

    def op_get_exc_value_addr(self, *args):
        raise NotImplementedError

    def op_instrument_count(self, ll_tag, ll_label):
        pass # xxx for now

    def op_keepalive(self, value):
        pass

    def op_hint(self, x, hints):
        return x

    def op_decode_arg(self, fname, i, name, vargs, vkwds):
        raise NotImplementedError("decode_arg")

    def op_decode_arg_def(self, fname, i, name, vargs, vkwds, default):
        raise NotImplementedError("decode_arg_def")

    def op_check_no_more_arg(self, fname, n, vargs):
        raise NotImplementedError("check_no_more_arg")

    def op_getslice(self, vargs, start, stop_should_be_None):
        raise NotImplementedError("getslice")   # only for argument parsing

    def op_check_self_nonzero(self, fname, vself):
        raise NotImplementedError("check_self_nonzero")

    def op_setfield(self, obj, fieldname, fieldvalue):
        # obj should be pointer
        FIELDTYPE = getattr(lltype.typeOf(obj).TO, fieldname)
        if FIELDTYPE is not lltype.Void:
            self.heap.setfield(obj, fieldname, fieldvalue)

    def op_bare_setfield(self, obj, fieldname, fieldvalue):
        # obj should be pointer
        FIELDTYPE = getattr(lltype.typeOf(obj).TO, fieldname)
        if FIELDTYPE is not lltype.Void:
            setattr(obj, fieldname, fieldvalue)

    def op_getinteriorfield(self, obj, *offsets):
        checkptr(obj)
        ob = obj
        for o in offsets:
            if isinstance(o, str):
                ob = getattr(ob, o)
            else:
                ob = ob[o]
        assert not isinstance(ob, lltype._interior_ptr)
        return ob

    def getinneraddr(self, obj, *offsets):
        TYPE = lltype.typeOf(obj).TO
        addr = llmemory.cast_ptr_to_adr(obj)
        for o in offsets:
            if isinstance(o, str):
                addr += llmemory.offsetof(TYPE, o)
                TYPE = getattr(TYPE, o)
            else:
                addr += llmemory.itemoffsetof(TYPE, o)
                TYPE = TYPE.OF
        return addr, TYPE

    def op_setinteriorfield(self, obj, *fieldnamesval):
        offsets, fieldvalue = fieldnamesval[:-1], fieldnamesval[-1]
        inneraddr, FIELD = self.getinneraddr(obj, *offsets)
        if FIELD is not lltype.Void:
            self.heap.setinterior(obj, inneraddr, FIELD, fieldvalue, offsets)

    def op_bare_setinteriorfield(self, obj, *fieldnamesval):
        offsets, fieldvalue = fieldnamesval[:-1], fieldnamesval[-1]
        inneraddr, FIELD = self.getinneraddr(obj, *offsets)
        if FIELD is not lltype.Void:
            llheap.setinterior(obj, inneraddr, FIELD, fieldvalue)

    def op_getarrayitem(self, array, index):
        return array[index]

    def op_setarrayitem(self, array, index, item):
        # array should be a pointer
        ITEMTYPE = lltype.typeOf(array).TO.OF
        if ITEMTYPE is not lltype.Void:
            self.heap.setarrayitem(array, index, item)

    def op_bare_setarrayitem(self, array, index, item):
        # array should be a pointer
        ITEMTYPE = lltype.typeOf(array).TO.OF
        if ITEMTYPE is not lltype.Void:
            array[index] = item

    def perform_call(self, f, ARGS, args):
        fobj = f._obj
        has_callable = getattr(fobj, '_callable', None) is not None
        if hasattr(fobj, 'graph'):
            graph = fobj.graph
        else:
            assert has_callable, "don't know how to execute %r" % f
            return self.invoke_callable_with_pyexceptions(f, *args)
        args_v = graph.getargs()
        if len(ARGS) != len(args_v):
            raise TypeError("graph with %d args called with wrong func ptr type: %r" %(len(args_v), ARGS))
        for T, v in zip(ARGS, args_v):
            if not lltype.isCompatibleType(T, v.concretetype):
                raise TypeError("graph with %r args called with wrong func ptr type: %r" %
                                (tuple([v.concretetype for v in args_v]), ARGS))
        frame = self.newsubframe(graph, args)
        return frame.eval()

    def op_direct_call(self, f, *args):
        FTYPE = lltype.typeOf(f).TO
        return self.perform_call(f, FTYPE.ARGS, args)

    def op_indirect_call(self, f, *args):
        graphs = args[-1]
        args = args[:-1]
        if graphs is not None:
            obj = f._obj
            if hasattr(obj, 'graph'):
                assert obj.graph in graphs
        else:
            pass
            #log.warn("op_indirect_call with graphs=None:", f)
        return self.op_direct_call(f, *args)

    def op_malloc(self, obj, flags):
        flavor = flags['flavor']
        zero = flags.get('zero', False)
        track_allocation = flags.get('track_allocation', True)
        if flavor == "stack":
            result = self.heap.malloc(obj, zero=zero, flavor='raw')
            self.alloca_objects.append(result)
            return result
        ptr = self.heap.malloc(obj, zero=zero, flavor=flavor,
                               track_allocation=track_allocation)
        return ptr

    def op_malloc_varsize(self, obj, flags, size):
        flavor = flags['flavor']
        zero = flags.get('zero', False)
        track_allocation = flags.get('track_allocation', True)
        assert flavor in ('gc', 'raw')
        try:
            ptr = self.heap.malloc(obj, size, zero=zero, flavor=flavor,
                                   track_allocation=track_allocation)
            return ptr
        except MemoryError:
            self.make_llexception()

    def op_free(self, obj, flags):
        assert flags['flavor'] == 'raw'
        track_allocation = flags.get('track_allocation', True)
        self.heap.free(obj, flavor='raw', track_allocation=track_allocation)

    def op_gc_add_memory_pressure(self, size):
        self.heap.add_memory_pressure(size)

    def op_gc_fq_next_dead(self, fq_tag):
        return self.heap.gc_fq_next_dead(fq_tag)

    def op_gc_fq_register(self, fq_tag, obj):
        self.heap.gc_fq_register(fq_tag, obj)

    def op_gc_gettypeid(self, obj):
        return lloperation.llop.combine_ushort(lltype.Signed, self.heap.gettypeid(obj), 0)

    def op_shrink_array(self, obj, smallersize):
        return self.heap.shrink_array(obj, smallersize)

    def op_zero_gc_pointers_inside(self, obj):
        raise NotImplementedError("zero_gc_pointers_inside")

    def op_gc_get_stats(self, obj):
        raise NotImplementedError("gc_get_stats")

    def op_gc_writebarrier_before_copy(self, source, dest,
                                       source_start, dest_start, length):
        if hasattr(self.heap, 'writebarrier_before_copy'):
            return self.heap.writebarrier_before_copy(source, dest,
                                                      source_start, dest_start,
                                                      length)
        else:
            return True

    def op_gc_writebarrier_before_move(self, array):
        if hasattr(self.heap, 'writebarrier_before_move'):
            self.heap.writebarrier_before_move(array)

    def op_getfield(self, obj, field):
        checkptr(obj)
        # check the difference between op_getfield and op_getsubstruct:
        assert not isinstance(getattr(lltype.typeOf(obj).TO, field),
                              lltype.ContainerType)
        return getattr(obj, field)

    def op_force_cast(self, RESTYPE, obj):
        from rpython.rtyper.lltypesystem import ll2ctypes
        return ll2ctypes.force_cast(RESTYPE, obj)
    op_force_cast.need_result_type = True

    def op_cast_int_to_ptr(self, RESTYPE, int1):
        return lltype.cast_int_to_ptr(RESTYPE, int1)
    op_cast_int_to_ptr.need_result_type = True

    def op_cast_ptr_to_int(self, ptr1):
        checkptr(ptr1)
        return lltype.cast_ptr_to_int(ptr1)

    def op_cast_opaque_ptr(self, RESTYPE, obj):
        checkptr(obj)
        return lltype.cast_opaque_ptr(RESTYPE, obj)
    op_cast_opaque_ptr.need_result_type = True

    def op_length_of_simple_gcarray_from_opaque(self, obj):
        checkptr(obj)
        return lltype.length_of_simple_gcarray_from_opaque(obj)

    def op_cast_ptr_to_adr(self, ptr):
        checkptr(ptr)
        return llmemory.cast_ptr_to_adr(ptr)

    def op_cast_adr_to_int(self, adr, mode):
        checkadr(adr)
        return llmemory.cast_adr_to_int(adr, mode)

    def op_convert_float_bytes_to_longlong(self, f):
        from rpython.rlib import longlong2float
        return longlong2float.float2longlong(f)

    def op_weakref_create(self, v_obj):
        def objgetter():    # special support for gcwrapper.py
            return self.getval(v_obj)
        assert self.llinterpreter.typer.getconfig().translation.rweakref
        return self.heap.weakref_create_getlazy(objgetter)
    op_weakref_create.specialform = True

    def op_weakref_deref(self, PTRTYPE, obj):
        assert self.llinterpreter.typer.getconfig().translation.rweakref
        return self.heap.weakref_deref(PTRTYPE, obj)
    op_weakref_deref.need_result_type = True

    def op_cast_ptr_to_weakrefptr(self, obj):
        assert self.llinterpreter.typer.getconfig().translation.rweakref
        return llmemory.cast_ptr_to_weakrefptr(obj)

    def op_cast_weakrefptr_to_ptr(self, PTRTYPE, obj):
        assert self.llinterpreter.typer.getconfig().translation.rweakref
        return llmemory.cast_weakrefptr_to_ptr(PTRTYPE, obj)
    op_cast_weakrefptr_to_ptr.need_result_type = True

    def op_gc__collect(self, *gen):
        self.heap.collect(*gen)

    def op_gc__collect_step(self):
        return self.heap.collect_step()

    def op_gc__enable(self):
        self.heap.enable()

    def op_gc__disable(self):
        self.heap.disable()

    def op_gc__isenabled(self):
        return self.heap.isenabled()

    def op_gc_heap_stats(self):
        raise NotImplementedError

    def op_gc_obtain_free_space(self, size):
        raise NotImplementedError

    def op_gc_can_move(self, ptr):
        addr = llmemory.cast_ptr_to_adr(ptr)
        return self.heap.can_move(addr)

    def op_gc_thread_run(self):
        self.heap.thread_run()

    def op_gc_thread_start(self):
        self.heap.thread_start()

    def op_gc_thread_die(self):
        self.heap.thread_die()

    def op_gc_thread_before_fork(self):
        raise NotImplementedError

    def op_gc_thread_after_fork(self):
        raise NotImplementedError

    def op_gc_free(self, addr):
        # what can you do?
        pass
        #raise NotImplementedError("gc_free")

    def op_gc_fetch_exception(self):
        raise NotImplementedError("gc_fetch_exception")

    def op_gc_restore_exception(self, exc):
        raise NotImplementedError("gc_restore_exception")

    def op_gc_adr_of_nursery_top(self):
        raise NotImplementedError
    def op_gc_adr_of_nursery_free(self):
        raise NotImplementedError

    def op_gc_adr_of_root_stack_base(self):
        raise NotImplementedError
    def op_gc_adr_of_root_stack_top(self):
        raise NotImplementedError

    def op_gc_modified_shadowstack(self):
        raise NotImplementedError

    def op_gc_call_rtti_destructor(self, rtti, addr):
        if hasattr(rtti._obj, 'destructor_funcptr'):
            d = rtti._obj.destructor_funcptr
            obptr = addr.ref()
            return self.op_direct_call(d, obptr)

    def op_gc_deallocate(self, TYPE, addr):
        raise NotImplementedError("gc_deallocate")

    def op_gc_reload_possibly_moved(self, v_newaddr, v_ptr):
        assert v_newaddr.concretetype is llmemory.Address
        assert isinstance(v_ptr.concretetype, lltype.Ptr)
        assert v_ptr.concretetype.TO._gckind == 'gc'
        newaddr = self.getval(v_newaddr)
        p = llmemory.cast_adr_to_ptr(newaddr, v_ptr.concretetype)
        if isinstance(v_ptr, Constant):
            assert v_ptr.value == p
        else:
            self.setvar(v_ptr, p)
    op_gc_reload_possibly_moved.specialform = True

    def op_gc_identityhash(self, obj):
        return lltype.identityhash(obj)

    def op_gc_id(self, ptr):
        PTR = lltype.typeOf(ptr)
        if isinstance(PTR, lltype.Ptr):
            return self.heap.gc_id(ptr)
        raise NotImplementedError("gc_id on %r" % (PTR,))

    def op_gc_set_max_heap_size(self, maxsize):
        raise NotImplementedError("gc_set_max_heap_size")

    def op_gc_stack_bottom(self):
        # Marker when we enter RPython code from C code.  It used to be
        # essential for trackgcroot.py.  Nowaways it is mostly unused,
        # except by revdb.
        pass

    def op_gc_pin(self, obj):
        addr = llmemory.cast_ptr_to_adr(obj)
        return self.heap.pin(addr)

    def op_gc_unpin(self, obj):
        addr = llmemory.cast_ptr_to_adr(obj)
        self.heap.unpin(addr)

    def op_gc__is_pinned(self, obj):
        addr = llmemory.cast_ptr_to_adr(obj)
        return self.heap._is_pinned(addr)

    def op_gc_get_type_info_group(self):
        raise NotImplementedError("gc_get_type_info_group")

    def op_gc_get_rpy_memory_usage(self):
        raise NotImplementedError("gc_get_rpy_memory_usage")

    def op_gc_get_rpy_roots(self):
        raise NotImplementedError("gc_get_rpy_roots")

    def op_gc_get_rpy_referents(self):
        raise NotImplementedError("gc_get_rpy_referents")

    def op_gc_is_rpy_instance(self):
        raise NotImplementedError("gc_is_rpy_instance")

    def op_gc_get_rpy_type_index(self):
        raise NotImplementedError("gc_get_rpy_type_index")

    def op_gc_dump_rpy_heap(self):
        raise NotImplementedError("gc_dump_rpy_heap")

    def op_gc_typeids_z(self):
        raise NotImplementedError("gc_typeids_z")

    def op_gc_typeids_list(self):
        raise NotImplementedError("gc_typeids_list")

    def op_gc_gcflag_extra(self, subopnum, *args):
        return self.heap.gcflag_extra(subopnum, *args)

    def op_gc_rawrefcount_init(self, *args):
        raise NotImplementedError("gc_rawrefcount_init")

    def op_gc_rawrefcount_to_obj(self, *args):
        raise NotImplementedError("gc_rawrefcount_to_obj")

    def op_gc_rawrefcount_from_obj(self, *args):
        raise NotImplementedError("gc_rawrefcount_from_obj")

    def op_gc_rawrefcount_create_link_pyobj(self, *args):
        raise NotImplementedError("gc_rawrefcount_create_link_pyobj")

    def op_gc_rawrefcount_create_link_pypy(self, *args):
        raise NotImplementedError("gc_rawrefcount_create_link_pypy")

    def op_gc_rawrefcount_mark_deallocating(self, *args):
        raise NotImplementedError("gc_rawrefcount_mark_deallocating")

    def op_gc_rawrefcount_next_dead(self, *args):
        raise NotImplementedError("gc_rawrefcount_next_dead")

    def op_do_malloc_fixedsize(self):
        raise NotImplementedError("do_malloc_fixedsize")
    def op_do_malloc_fixedsize_clear(self):
        raise NotImplementedError("do_malloc_fixedsize_clear")
    def op_do_malloc_varsize(self):
        raise NotImplementedError("do_malloc_varsize")
    def op_do_malloc_varsize_clear(self):
        raise NotImplementedError("do_malloc_varsize_clear")

    def op_get_write_barrier_failing_case(self):
        raise NotImplementedError("get_write_barrier_failing_case")

    def op_get_write_barrier_from_array_failing_case(self):
        raise NotImplementedError("get_write_barrier_from_array_failing_case")

    def op_stack_current(self):
        return 0

    def op_threadlocalref_addr(self):
        return _address_of_thread_local()

    def op_threadlocalref_get(self, RESTYPE, offset):
        return self.op_raw_load(RESTYPE, _address_of_thread_local(), offset)
    op_threadlocalref_get.need_result_type = True

    op_threadlocalref_load = op_threadlocalref_get

    def op_threadlocalref_store(self, offset, value):
        self.op_raw_store(_address_of_thread_local(), offset, value)

    def op_threadlocalref_acquire(self, prev):
        raise NotImplementedError
    def op_threadlocalref_release(self, prev):
        raise NotImplementedError
    def op_threadlocalref_enum(self, prev):
        raise NotImplementedError

    # __________________________________________________________
    # operations on addresses

    def op_raw_malloc(self, size, zero):
        assert lltype.typeOf(size) == lltype.Signed
        return llmemory.raw_malloc(size, zero=zero)

    def op_boehm_malloc(self, size):
        assert lltype.typeOf(size) == lltype.Signed
        raw = llmemory.raw_malloc(size)
        return llmemory.cast_adr_to_ptr(raw, llmemory.GCREF)
    op_boehm_malloc_atomic = op_boehm_malloc

    def op_boehm_register_finalizer(self, p, finalizer):
        pass

    def op_boehm_disappearing_link(self, link, obj):
        pass

    def op_raw_malloc_usage(self, size):
        assert lltype.typeOf(size) == lltype.Signed
        return llmemory.raw_malloc_usage(size)

    def op_raw_free(self, addr):
        checkadr(addr)
        llmemory.raw_free(addr)

    def op_raw_memclear(self, addr, size):
        checkadr(addr)
        llmemory.raw_memclear(addr, size)

    def op_raw_memcopy(self, fromaddr, toaddr, size):
        checkadr(fromaddr)
        checkadr(toaddr)
        llmemory.raw_memcopy(fromaddr, toaddr, size)

    def op_raw_memset(self, addr, byte, size):
        raise NotImplementedError

    op_raw_memmove = op_raw_memcopy # this is essentially the same here

    def op_raw_load(self, RESTYPE, addr, offset):
        checkadr(addr)
        if isinstance(offset, int):
            from rpython.rtyper.lltypesystem import rffi
            ll_p = rffi.cast(rffi.CCHARP, addr)
            ll_p = rffi.cast(rffi.CArrayPtr(RESTYPE),
                             rffi.ptradd(ll_p, offset))
            value = ll_p[0]
        elif getattr(addr, 'is_fake_thread_local_addr', False):
            assert type(offset) is CDefinedIntSymbolic
            value = self.llinterpreter.get_tlobj()[offset.expr]
        else:
            assert offset.TYPE == RESTYPE
            value = getattr(addr, str(RESTYPE).lower())[offset.repeat]
        assert lltype.typeOf(value) == RESTYPE
        return value
    op_raw_load.need_result_type = True

    def op_raw_store(self, addr, offset, value):
        # XXX handle the write barrier by delegating to self.heap instead
        self.op_bare_raw_store(addr, offset, value)

    def op_bare_raw_store(self, addr, offset, value):
        checkadr(addr)
        ARGTYPE = lltype.typeOf(value)
        if isinstance(offset, int):
            from rpython.rtyper.lltypesystem import rffi
            ll_p = rffi.cast(rffi.CCHARP, addr)
            ll_p = rffi.cast(rffi.CArrayPtr(ARGTYPE),
                             rffi.ptradd(ll_p, offset))
            ll_p[0] = value
        elif getattr(addr, 'is_fake_thread_local_addr', False):
            assert type(offset) is CDefinedIntSymbolic
            self.llinterpreter.get_tlobj()[offset.expr] = value
        elif isinstance(offset, llmemory.ArrayLengthOffset):
            assert len(addr.ptr) == value  # invalid ArrayLengthOffset
        else:
            assert offset.TYPE == ARGTYPE
            getattr(addr, str(ARGTYPE).lower())[offset.repeat] = value

    def op_track_alloc_start(self, addr):
        # we don't do tracking at this level
        checkadr(addr)

    def op_track_alloc_stop(self, addr):
        checkadr(addr)

    def op_gc_enter_roots_frame(self, gcdata, numcolors):
        """Fetch from the gcdata the current root_stack_top; bump it
        by 'numcolors'; and assert that the new area is fully
        uninitialized so far.
        """
        assert not hasattr(self, '_inside_roots_frame')
        p = gcdata.inst_root_stack_top.ptr
        q = lltype.direct_ptradd(p, numcolors)
        self._inside_roots_frame = (p, q, numcolors, gcdata)
        gcdata.inst_root_stack_top = llmemory.cast_ptr_to_adr(q)
        #
        array = p._obj._parentstructure()
        index = p._obj._parent_index
        for i in range(index, index + numcolors):
            assert isinstance(array.getitem(i), lltype._uninitialized)

    def op_gc_leave_roots_frame(self):
        """Cancel gc_enter_roots_frame() by removing the frame from
        the root_stack_top.  Writes uninitialized entries in its old place.
        """
        (p, q, numcolors, gcdata) = self._inside_roots_frame
        assert gcdata.inst_root_stack_top.ptr == q
        gcdata.inst_root_stack_top = llmemory.cast_ptr_to_adr(p)
        del self._inside_roots_frame
        #
        array = p._obj._parentstructure()
        index = p._obj._parent_index
        for i in range(index, index + numcolors):
            array.setitem(i, lltype._uninitialized(llmemory.Address))

    def op_gc_save_root(self, num, value):
        """Save one value (int or ptr) into the frame."""
        (p, q, numcolors, gcdata) = self._inside_roots_frame
        assert 0 <= num < numcolors
        if isinstance(value, int):
            assert value & 1    # must be odd
            v = llmemory.cast_int_to_adr(value)
        else:
            v = llmemory.cast_ptr_to_adr(value)
        llmemory.cast_ptr_to_adr(p).address[num] = v

    def op_gc_restore_root(self, c_num, v_value):
        """Restore one value from the frame."""
        num = c_num.value
        (p, q, numcolors, gcdata) = self._inside_roots_frame
        assert 0 <= num < numcolors
        assert isinstance(v_value.concretetype, lltype.Ptr)
        assert v_value.concretetype.TO._gckind == 'gc'
        newvalue = llmemory.cast_ptr_to_adr(p).address[num]
        newvalue = llmemory.cast_adr_to_ptr(newvalue, v_value.concretetype)
        self.setvar(v_value, newvalue)
    op_gc_restore_root.specialform = True

    def op_gc_push_roots(self, *args):
        raise NotImplementedError

    def op_gc_pop_roots(self, *args):
        raise NotImplementedError

    # ____________________________________________________________
    # Overflow-detecting variants

    def op_int_add_ovf(self, x, y):
        assert isinstance(x, (int, long, llmemory.AddressOffset))
        assert isinstance(y, (int, long, llmemory.AddressOffset))
        try:
            return ovfcheck(x + y)
        except OverflowError:
            self.make_llexception()

    def op_int_add_nonneg_ovf(self, x, y):
        if isinstance(y, int):
            assert y >= 0
        return self.op_int_add_ovf(x, y)

    def op_int_sub_ovf(self, x, y):
        assert isinstance(x, (int, long))
        assert isinstance(y, (int, long))
        try:
            return ovfcheck(x - y)
        except OverflowError:
            self.make_llexception()

    def op_int_mul_ovf(self, x, y):
        assert isinstance(x, (int, long, llmemory.AddressOffset))
        assert isinstance(y, (int, long, llmemory.AddressOffset))
        try:
            return ovfcheck(x * y)
        except OverflowError:
            self.make_llexception()

    def op_int_is_true(self, x):
        # special case
        if type(x) is CDefinedIntSymbolic:
            x = x.default
        # if type(x) is a subclass of Symbolic, bool(x) will usually raise
        # a TypeError -- unless __nonzero__ has been explicitly overridden.
        assert is_valid_int(x) or isinstance(x, Symbolic)
        return bool(x)

    # hack for jit.codegen.llgraph

    def op_check_and_clear_exc(self):
        exc_data = self.llinterpreter.get_transformed_exc_data(self.graph)
        assert exc_data
        etype = exc_data.exc_type
        evalue = exc_data.exc_value
        exc_data.exc_type = lltype.typeOf(etype)._defl()
        exc_data.exc_value = lltype.typeOf(evalue)._defl()
        return bool(etype)

    def op_gc_move_out_of_nursery(self, obj):
        raise NotImplementedError("gc_move_out_of_nursery")

    def op_gc_increase_root_stack_depth(self, new_depth):
        raise NotImplementedError("gc_increase_root_stack_depth")

    def op_revdb_stop_point(self, *args):
        pass
    def op_revdb_send_answer(self, *args):
        raise NotImplementedError
    def op_revdb_breakpoint(self, *args):
        raise NotImplementedError
    def op_revdb_get_value(self, *args):
        raise NotImplementedError
    def op_revdb_get_unique_id(self, *args):
        raise NotImplementedError
    def op_revdb_watch_save_state(self, *args):
        return False
    def op_revdb_watch_restore_state(self, *args):
        raise NotImplementedError
    def op_revdb_weakref_create(self, *args):
        raise NotImplementedError
    def op_revdb_weakref_deref(self, *args):
        raise NotImplementedError
    def op_revdb_call_destructor(self, *args):
        raise NotImplementedError
    def op_revdb_strtod(self, *args):
        raise NotImplementedError
    def op_revdb_frexp(self, *args):
        raise NotImplementedError
    def op_revdb_modf(self, *args):
        raise NotImplementedError
    def op_revdb_dtoa(self, *args):
        raise NotImplementedError
    def op_revdb_do_next_call(self):
        pass
    def op_revdb_set_thread_breakpoint(self, *args):
        raise NotImplementedError


class Tracer(object):
    Counter = 0
    file = None
    TRACE = int(os.getenv('PYPY_TRACE') or '0')

    HEADER = """<html><head>
        <script language=javascript type='text/javascript'>
        function togglestate(n) {
          var item = document.getElementById('div'+n)
          if (item.style.display == 'none')
            item.style.display = 'block';
          else
            item.style.display = 'none';
        }

        function toggleall(lst) {
          for (var i = 0; i<lst.length; i++) {
            togglestate(lst[i]);
          }
        }
        </script>
        </head>

        <body><pre>
    """

    FOOTER = """</pre>
        <script language=javascript type='text/javascript'>
        toggleall(%r);
        </script>

    </body></html>"""

    ENTER = ('''\n\t<a href="javascript:togglestate(%d)">%s</a>'''
             '''\n<div id="div%d" style="display: %s">\t''')
    LEAVE = '''\n</div>\t'''

    def htmlquote(self, s, text_to_html={}):
        # HTML quoting, lazily initialized
        if not text_to_html:
            import htmlentitydefs
            for key, value in htmlentitydefs.entitydefs.items():
                text_to_html[value] = '&' + key + ';'
        return ''.join([text_to_html.get(c, c) for c in s])

    def start(self):
        # start of a dump file
        if not self.TRACE:
            return
        from rpython.tool.udir import udir
        n = Tracer.Counter
        Tracer.Counter += 1
        filename = 'llinterp_trace_%d.html' % n
        self.file = udir.join(filename).open('w')
        print(self.HEADER, file=self.file)

        linkname = str(udir.join('llinterp_trace.html'))
        try:
            os.unlink(linkname)
        except OSError:
            pass
        try:
            os.symlink(filename, linkname)
        except (AttributeError, OSError):
            pass

        self.count = 0
        self.indentation = ''
        self.depth = 0
        self.latest_call_chain = []

    def stop(self):
        # end of a dump file
        if self.file:
            print(self.FOOTER % (self.latest_call_chain[1:]), file=self.file)
            self.file.close()
            self.file = None

    def enter(self, graph):
        # enter evaluation of a graph
        if self.file:
            del self.latest_call_chain[self.depth:]
            self.depth += 1
            self.latest_call_chain.append(self.count)
            s = self.htmlquote(str(graph))
            i = s.rfind(')')
            s = s[:i+1] + '<b>' + s[i+1:] + '</b>'
            if self.count == 0:
                display = 'block'
            else:
                display = 'none'
            text = self.ENTER % (self.count, s, self.count, display)
            self.indentation += '    '
            self.file.write(text.replace('\t', self.indentation))
            self.count += 1

    def leave(self):
        # leave evaluation of a graph
        if self.file:
            self.indentation = self.indentation[:-4]
            self.file.write(self.LEAVE.replace('\t', self.indentation))
            self.depth -= 1

    def dump(self, text, bold=False):
        if self.file:
            text = self.htmlquote(text)
            if bold:
                text = '<b>%s</b>' % (text,)
            self.file.write(text.replace('\n', '\n'+self.indentation))

    def flush(self):
        if self.file:
            self.file.flush()

def wrap_callable(llinterpreter, fn, obj, method_name):
    if method_name is None:
        # fn is a StaticMethod
        if obj is not None:
            self_arg = [obj]
        else:
            self_arg = []
        func_graph = fn.graph
    else:
        # obj is an instance, we want to call 'method_name' on it
        assert fn is None
        self_arg = [obj]
        func_graph = obj._TYPE._methods[method_name._str].graph

    return wrap_graph(llinterpreter, func_graph, self_arg)

def wrap_graph(llinterpreter, graph, self_arg):
    """
    Returns a callable that inteprets the given func or method_name when called.
    """

    def interp_func(*args):
        graph_args = self_arg + list(args)
        return llinterpreter.eval_graph(graph, args=graph_args)
    interp_func.graph = graph
    interp_func.self_arg = self_arg
    return graph.name, interp_func


def enumerate_exceptions_top_down():
    import exceptions
    result = []
    seen = {}
    def addcls(cls):
        if (type(cls) is type(Exception) and
            issubclass(cls, py.builtin.BaseException)):
            if cls in seen:
                return
            for base in cls.__bases__:   # bases first
                addcls(base)
            result.append(cls)
            seen[cls] = True
    for cls in exceptions.__dict__.values():
        addcls(cls)
    return result

class _address_of_local_var(object):
    _TYPE = llmemory.Address
    def __init__(self, frame, v):
        self._frame = frame
        self._v = v
    def _getaddress(self):
        return _address_of_local_var_accessor(self._frame, self._v)
    address = property(_getaddress)

class _address_of_local_var_accessor(object):
    def __init__(self, frame, v):
        self.frame = frame
        self.v = v
    def __getitem__(self, index):
        if index != 0:
            raise IndexError("address of local vars only support [0] indexing")
        p = self.frame.getval(self.v)
        result = llmemory.cast_ptr_to_adr(p)
        # the GC should never see instances of _gctransformed_wref
        result = self.unwrap_possible_weakref(result)
        return result
    def __setitem__(self, index, newvalue):
        if index != 0:
            raise IndexError("address of local vars only support [0] indexing")
        if self.v.concretetype == llmemory.WeakRefPtr:
            # fish some more
            assert isinstance(newvalue, llmemory.fakeaddress)
            p = llmemory.cast_ptr_to_weakrefptr(newvalue.ptr)
        else:
            p = llmemory.cast_adr_to_ptr(newvalue, self.v.concretetype)
        self.frame.setvar(self.v, p)
    def unwrap_possible_weakref(self, addr):
        # fish fish fish
        if addr and isinstance(addr.ptr._obj, llmemory._gctransformed_wref):
            return llmemory.fakeaddress(addr.ptr._obj._ptr)
        return addr

class _address_of_thread_local(object):
    _TYPE = llmemory.Address
    is_fake_thread_local_addr = True



import py
from rpython.jit.backend.arm.test.support import JitARMMixin
from rpython.jit.metainterp.test import test_loop_unroll

class TestLoopSpec(JitARMMixin, test_loop_unroll.LoopUnrollTest):
    # for the individual tests see
    # ====> ../../../metainterp/test/test_loop.py
    pass



#!/usr/bin/env python
""" Parse gcdumps. Use by saying PYPYLOG=gc-collect:log pypy <program>
and run it by:

gcanalyze.py logfile [--plot]
"""
from __future__ import print_function

import sys
from rpython.tool.logparser import parse_log

NO_BUCKETS = 8

def main(arg, plot=False):
    log = parse_log(open(arg).readlines())
    all = []
    for entry in log:
        if entry[0].startswith('gc-collect'):
            start = entry[1]
            end = entry[2]
            all.append(float(end - start) / 1000000)
    format_output(all, plot=plot)

def format_output(all, plot=False):
    avg = sum(all) / len(all)
    max_t = max(all)
    print("AVG:", "%.1fms" % avg, "MAX:", "%.1fms" % max_t, "TOTAL:" , "%.1fms" % sum(all))
    print()
    buckets = [0] * NO_BUCKETS
    for item in all:
        bucket = int(item / max_t * NO_BUCKETS)
        if bucket == len(buckets):
            bucket = len(buckets) - 1
        buckets[bucket] += 1
    l1 = ["%.1fms" % ((i + 1) * max_t / NO_BUCKETS) for i in range(NO_BUCKETS)]
    l2 = [str(i) for i in buckets]
    for i, elem in enumerate(l1):
        l2[i] += " " * (len(elem) - len(l2[i]))
    if plot:
        l3 = ["+"]
        for i, elem in enumerate(l1):
            l3.append("-" * len(elem) + "+")
        l1.insert(0, "")
        l2.insert(0, "")
        print("".join(l3))
        print("|".join(l1) + "|")
        print("".join(l3))
        print("|".join(l2) + "|")
        print("".join(l3))
    else:
        print(" ".join(l1))
        print(" ".join(l2))

if __name__ == '__main__':
    if len(sys.argv) < 2 or len(sys.argv) > 3:
        print(__doc__)
        sys.exit(1)
    plot = False
    if len(sys.argv) == 3:
        if sys.argv[1] == '--plot':
            plot = True
            arg = sys.argv[2]
        elif sys.argv[2] == '--plot':
            plot = True
            arg = sys.argv[1]
        else:
            print("Wrong command line options:", sys.argv)
            sys.exit(1)
    else:
        arg = sys.argv[1]
    main(arg, plot=plot)



""" transparent list implementation
"""
from pypy.interpreter import baseobjspace
from pypy.interpreter.error import OperationError, oefmt


def transparent_class(name, BaseCls):
    class W_Transparent(BaseCls):
        def __init__(self, space, w_type, w_controller):
            self.w_type = w_type
            self.w_controller = w_controller

        def descr_call_mismatch(self, space, name, reqcls, args):
            args_w = args.arguments_w[:]
            args_w[0] = space.newtext(name)
            args = args.replace_arguments(args_w)
            return space.call_args(self.w_controller, args)

        def getclass(self, space):
            return self.w_type

        def setclass(self, space, w_subtype):
            raise oefmt(space.w_TypeError,
                        "You cannot override __class__ for transparent "
                        "proxies")

        def getdictvalue(self, space, attr):
            try:
                return space.call_function(self.w_controller, space.newtext('__getattribute__'),
                   space.newtext(attr))
            except OperationError as e:
                if not e.match(space, space.w_AttributeError):
                    raise
                return None

        def setdictvalue(self, space, attr, w_value):
            try:
                space.call_function(self.w_controller, space.newtext('__setattr__'),
                   space.newtext(attr), w_value)
                return True
            except OperationError as e:
                if not e.match(space, space.w_AttributeError):
                    raise
                return False

        def deldictvalue(self, space, attr):
            try:
                space.call_function(self.w_controller, space.newtext('__delattr__'),
                   space.newtext(attr))
                return True
            except OperationError as e:
                if not e.match(space, space.w_AttributeError):
                    raise
                return False

        def getdict(self, space):
            return self.getdictvalue(space, '__dict__')

        def setdict(self, space, w_dict):
            if not self.setdictvalue(space, '__dict__', w_dict):
                baseobjspace.W_Root.setdict(self, space, w_dict)

    W_Transparent.__name__ = name
    return W_Transparent

W_Transparent = transparent_class('W_Transparent', baseobjspace.W_Root)



from rpython.rlib.parsing.lexer import *
from rpython.rlib.parsing.regex import *
from rpython.translator.c.test.test_genc import compile

def test_translate_simple():
    digits = RangeExpression("0", "9")
    lower = RangeExpression("a", "z")
    upper = RangeExpression("A", "Z")
    keywords = StringExpression("if") | StringExpression("else") | StringExpression("def") | StringExpression("class")
    underscore = StringExpression("_")
    atoms = lower + (upper | lower | digits | underscore).kleene()
    vars = underscore | (upper + (upper | lower | underscore | digits).kleene())
    integers = StringExpression("0") | (RangeExpression("1", "9") + digits.kleene())
    white = StringExpression(" ")
    token_regexes = [keywords, atoms, vars, integers, white]
    names = ["KEYWORD", "ATOM", "VAR", "INT", "WHITE"]
    l1 = Lexer(token_regexes, names, None)
    l2 = Lexer(token_regexes, names, ["WHITE"])

    def lex(s, ignore=False):
        if ignore:
            tokens = l2.tokenize(s)
        else:
            tokens = l1.tokenize(s)
        return "-%-".join([t.name for t in tokens])

    res = lex("if A a 12341 0 else").split("-%-")
    assert res == ("KEYWORD WHITE VAR WHITE ATOM WHITE INT WHITE "
                   "INT WHITE KEYWORD").split()
    res = lex("if A a 12341 0 else", True).split("-%-")
    assert res == "KEYWORD VAR ATOM INT INT KEYWORD".split()

    func = compile(lex, [str, bool])
    res = func("if A a 12341 0 else", False).split("-%-")
    assert res == ("KEYWORD WHITE VAR WHITE ATOM WHITE INT WHITE "
                   "INT WHITE KEYWORD").split()
    res = func("if A a 12341 0 else", True).split("-%-")
    assert res == "KEYWORD VAR ATOM INT INT KEYWORD".split()




import py, sys, re
from pypy.module.pypyjit.test_pypy_c.test_00_model import BaseTestPyPyC

class TestCProfile(BaseTestPyPyC):

    def test_cprofile_builtin(self):
        def main(n):
            import _lsprof
            prof = _lsprof.Profiler()
            i = 0
            lst = []
            prof.enable()
            while i < n:
                lst.append(i)   # ID: append
                lst.pop()       # ID: pop
                i += 1
            prof.disable()
            return [(entry.code, entry.callcount) for entry in prof.getstats()]
        #
        log = self.run(main, [500])
        assert sorted(log.result) == [
            ("<method 'append' of 'list' objects>", 500),
            ("<method 'disable' of '_lsprof.Profiler' objects>", 1),
            ("<method 'pop' of 'list' objects>", 500),
            ]
        for method in ['append', 'pop']:
            loop, = log.loops_by_id(method)
            print loop.ops_by_id(method)
            # on 32-bit, there is f1=call(read_timestamp); ...;
            # f2=call(read_timestamp); f3=call(llong_sub,f1,f2)
            # but all calls can be special-cased by the backend if
            # supported.  On 64-bit there is only the two calls to
            # read_timestamp.
            r = re.compile(r" call_\w[(]ConstClass[(](.+?)[)]")
            calls = r.findall(repr(loop.ops_by_id(method)))
            if sys.maxint == 2147483647:
                assert len(calls) == 6
            else:
                assert len(calls) == 2
            for x in calls:
                assert ('ll_read_timestamp' in x or 'llong_sub' in x
                        or 'llong_add' in x)
            #
            assert ' call_may_force(' not in repr(loop.ops_by_id(method))
            assert ' cond_call(' in repr(loop.ops_by_id(method))




def globals(space):
    "Return the dictionary containing the current scope's global variables."
    ec = space.getexecutioncontext()
    return ec.gettopframe_nohidden().get_w_globals()

def locals(space):
    """Return a dictionary containing the current scope's local variables.
Note that this may be the real dictionary of local variables, or a copy."""
    ec = space.getexecutioncontext()
    return ec.gettopframe_nohidden().getdictscope()




from __future__ import absolute_import

import math

import _numpypy
from _numpypy.multiarray import set_docstring

def arange(start, stop=None, step=1, dtype=None):
    '''arange([start], stop[, step], dtype=None)
    Generate values in the half-interval [start, stop).
    '''
    if stop is None:
        stop = start
        start = 0
    if dtype is None:
        # find minimal acceptable dtype but not less than int
        dtype = _numpypy.multiarray.result_type(start, stop, step, int)
    length = math.ceil((float(stop) - start) / step)
    length = int(length)
    arr = _numpypy.multiarray.empty(length, dtype=dtype)
    i = start
    for j in xrange(arr.size):
        arr[j] = i
        i += step
    return arr


def add_docstring(obj, docstring):
    old_doc = getattr(obj, '__doc__', None)
    if old_doc is not None:
        raise RuntimeError("%s already has a docstring" % obj)
    try:
        set_docstring(obj, docstring)
    except:
        raise TypeError("Cannot set a docstring for %s" % obj)



import sys
import os
from hashlib import md5

import z3

from rpython.jit.metainterp.optimizeopt.test.test_z3intbound import (
    make_z3_intbounds_instance,
)
from rpython.jit.metainterp.optimizeopt.intutils import IntBound
from rpython.jit.metainterp.ruleopt import parse
from rpython.rlib.rarithmetic import LONG_BIT, intmask, r_uint

from rpython.config.translationoption import CACHE_DIR
from rpython.tool.gcc_cache import try_atomic_write

MAXINT = sys.maxint
MININT = -sys.maxint - 1

class ProofProblem(Exception):
    pass

class CouldNotProve(ProofProblem):
    def __init__(self, rule, cond, model, lhs, rhs, prover):
        self.rule = rule
        self.cond = cond
        self.model = model
        self.rhs = rhs
        self.lhs = lhs
        self.prover = prover

    def format(self):
        rule = self.rule
        res = ["Could not prove correctness of rule '%s'" % self.rule.name]
        if self.rule.sourcepos:
            res.append("in line %s" % (self.rule.sourcepos.lineno, ))
        prover = self.prover
        model = prover.solver.model()
        detail = []
        res.append("counterexample given by Z3:")
        res.append("counterexample values:")
        for name, bound in prover.name_to_intbound.iteritems():
            if name in prover.glue_conditions_added:
                realbound = IntBound(model.evaluate(bound.lower).as_signed_long(),
                                      model.evaluate(bound.upper).as_signed_long(),
                                      r_uint(model.evaluate(bound.tmask).as_signed_long()),
                                      r_uint(model.evaluate(bound.tvalue).as_signed_long()),)
                detail.append("bounds for %s: %s" % (name, realbound))
            res.append("%s: %s" % (name, model[prover.name_to_z3[name]].as_signed_long()))
        res.append("operation %s with Z3 formula %s" % (rule.pattern, self.lhs))
        res.append("has counterexample result vale: %s" % (model.evaluate(self.lhs).as_signed_long(), ))
        res.append("BUT")
        res.append("target expression: %s with Z3 formula %s" % (rule.target, self.rhs))
        res.append("has counterexample value: %s" % (model.evaluate(self.rhs).as_signed_long(), ))
        res.extend(detail)
        return "\n".join(res)

class RuleCannotApply(ProofProblem):
    def __init__(self, rule, cond, prover):
        self.rule = rule
        self.cond = cond
        self.prover = prover

    def format(self):
        rule = self.rule
        res = ["Rule '%s' cannot ever apply" % self.rule.name]
        if self.rule.sourcepos:
            res.append("in line %s" % (self.rule.sourcepos.lineno, ))
        prover = self.prover
        res.append("Z3 did not manage to find values for variables %s such that the following condition becomes True:" % ", ".join(prover.name_to_z3))
        res.append(str(self.cond))
        return "\n".join(res)

TRUEBV = z3.BitVecVal(1, LONG_BIT)
FALSEBV = z3.BitVecVal(0, LONG_BIT)


def z3_cond(z3expr):
    return z3.If(z3expr, TRUEBV, FALSEBV)


def z3_bool_expression(opname, arg0, arg1=None):
    expr = None
    valid = True
    if opname == "int_eq":
        expr = arg0 == arg1
    elif opname == "int_ne":
        expr = arg0 != arg1
    elif opname == "int_lt":
        expr = arg0 < arg1
    elif opname == "int_le":
        expr = arg0 <= arg1
    elif opname == "int_gt":
        expr = arg0 > arg1
    elif opname == "int_ge":
        expr = arg0 >= arg1
    elif opname == "uint_lt":
        expr = z3.ULT(arg0, arg1)
    elif opname == "uint_le":
        expr = z3.ULE(arg0, arg1)
    elif opname == "uint_gt":
        expr = z3.UGT(arg0, arg1)
    elif opname == "uint_ge":
        expr = z3.UGE(arg0, arg1)
    elif opname == "int_is_true":
        expr = arg0 != FALSEBV
    elif opname == "int_is_zero":
        expr = arg0 == FALSEBV
    else:
        assert 0
    return expr, valid


def z3_expression(opname, arg0, arg1=None):
    expr = None
    valid = True
    if opname == "int_add":
        expr = arg0 + arg1
    elif opname == "int_sub":
        expr = arg0 - arg1
    elif opname == "int_mul":
        expr = arg0 * arg1
    elif opname == "int_and":
        expr = arg0 & arg1
    elif opname == "int_or":
        expr = arg0 | arg1
    elif opname == "int_xor":
        expr = arg0 ^ arg1
    elif opname == "int_lshift":
        expr = arg0 << arg1
        valid = z3.And(arg1 >= 0, arg1 < LONG_BIT)
    elif opname == "int_rshift":
        expr = arg0 >> arg1
        valid = z3.And(arg1 >= 0, arg1 < LONG_BIT)
    elif opname == "uint_rshift":
        expr = z3.LShR(arg0, arg1)
        valid = z3.And(arg1 >= 0, arg1 < LONG_BIT)
    elif opname == "uint_mul_high":
        # zero-extend args to 2*LONG_BIT bit, then multiply and extract
        # highest LONG_BIT bits
        zarg0 = z3.ZeroExt(LONG_BIT, arg0)
        zarg1 = z3.ZeroExt(LONG_BIT, arg1)
        expr = z3.Extract(LONG_BIT * 2 - 1, LONG_BIT, zarg0 * zarg1)
    elif opname == "int_neg":
        expr = -arg0
    elif opname == "int_invert":
        expr = ~arg0
    elif opname == "int_force_ge_zero":
        expr = z3.If(arg0 < 0, 0, arg0)
    else:
        expr, valid = z3_bool_expression(opname, arg0, arg1)
        return z3_cond(expr), valid
    return expr, valid


def z3_and(*args):
    args = [arg for arg in args if arg is not True]
    if args:
        if len(args) == 1:
            return args[0]
        return z3.And(*args)
    return True


def z3_implies(a, b):
    if a is True:
        return b
    return z3.Implies(a, b)


def popcount64(w):
    w -= (w >> 1) & 0x5555555555555555
    w = (w & 0x3333333333333333) + ((w >> 2) & 0x3333333333333333)
    w = (w + (w >> 4)) & 0x0F0F0F0F0F0F0F0F
    return ((w * 0x0101010101010101) >> 56) & 0xFF


def highest_bit(x):
    x |= x >> 1
    x |= x >> 2
    x |= x >> 4
    x |= x >> 8
    x |= x >> 16
    x |= x >> 32
    return popcount64(x) - 1


def z3_highest_bit(x):
    x |= z3.LShR(x, 1)
    x |= z3.LShR(x, 2)
    x |= z3.LShR(x, 4)
    x |= z3.LShR(x, 8)
    x |= z3.LShR(x, 16)
    x |= z3.LShR(x, 32)
    return popcount64(x) - 1

def z3_min(a, b):
    return z3.If(a <= b, a, b)

class Prover(parse.Visitor):
    def __init__(self):
        self.solver = z3.Optimize()
        self.name_to_z3 = {}
        self.name_to_intbound = {}
        self.glue_conditions_added = set()
        self.glue_conditions = []

    def prove(self, cond):
        z3res = self.solver.check(z3.Not(cond))
        if z3res == z3.unsat:
            return True
        elif z3res == z3.unknown:
            return False
        elif z3res == z3.sat:
            global model
            model = self.solver.model()
            return False

    def _convert_var(self, name):
        def newvar(name, suffix=""):
            if suffix:
                name += "_" + suffix
            res = z3.BitVec(name, LONG_BIT)
            self.name_to_z3[name] = res
            return res

        if name in self.name_to_z3:
            return self.name_to_z3[name]
        res = newvar(name)
        self.solver.minimize(res)
        b = make_z3_intbounds_instance(name, res)
        self.name_to_intbound[name] = b
        return res

    def _convert_intbound(self, name):
        b = self.name_to_intbound[name]
        if name not in self.glue_conditions_added:
            self.glue_conditions.append(b.z3_formula())
            self.glue_conditions_added.add(name)
        return b

    def _convert_attr(
        self,
        varname,
        attrname,
    ):
        b = self._convert_intbound(varname)
        if attrname == 'ones':
            return b.tvalue
        if attrname == 'zeros':
            return ~(b.tvalue | b.tmask)
        return getattr(b, attrname)

    def visit_PatternOp(self, pattern):
        args = [self.visit(arg) for arg in pattern.args]
        res, valid = z3_expression(pattern.opname, *[arg[0] for arg in args])
        return res, z3_and(valid, *[arg[1] for arg in args])

    def visit_PatternVar(self, pattern):
        return self._convert_var(pattern.name), True

    def visit_PatternConst(self, pattern):
        if pattern.const == "LONG_BIT":
            return z3.BitVecVal(LONG_BIT, LONG_BIT)
        elif pattern.const == "MININT":
            return z3.BitVecVal(MININT, LONG_BIT), True
        elif pattern.const == "MAXINT":
            return z3.BitVecVal(MAXINT, LONG_BIT), True
        res = z3.BitVecVal(pattern.const, LONG_BIT)
        return res, True

    def visit_ShortcutOr(self, expr, targettype=int):
        assert targettype is bool
        left, leftvalid = self.visit(expr.left, bool)
        right, rightvalid = self.visit(expr.right, bool)
        res = z3.If(left, left, right)
        return res, z3_and(leftvalid, rightvalid)

    def visit_ShortcutAnd(self, expr, targettype=int):
        assert targettype is bool
        left, leftvalid = self.visit(expr.left, bool)
        right, rightvalid = self.visit(expr.right, bool)
        res = z3.If(left, right, left)
        return res, z3_and(leftvalid, rightvalid)

    def visit_BinOp(self, expr, targettype=int):
        left, leftvalid = self.visit(expr.left, int)
        right, rightvalid = self.visit(expr.right, int)
        if targettype is int:
            res, valid = z3_expression(expr.opname, left, right)
        else:
            assert targettype is bool
            res, valid = z3_bool_expression(expr.opname, left, right)
        return res, z3_and(leftvalid, rightvalid, valid)

    def visit_UnaryOp(self, expr, targettype=int):
        assert targettype is int
        left, leftvalid = self.visit(expr.left, targettype)
        res, valid = z3_expression(expr.opname, left)
        return res, z3_and(leftvalid, valid)

    def visit_Name(self, expr, targettype=int):
        if expr.name == "LONG_BIT":
            return 64, True
        if expr.name == "MAXINT":
            return MAXINT, True
        if expr.name == "MININT":
            import pdb;pdb.set_trace()
            return MININT, True
        if targettype is int:
            var = self._convert_var(expr.name)
            return var, True
        if targettype is IntBound:
            b = self._convert_intbound(expr.name)
            return b, True
        import pdb

        pdb.set_trace()

    def visit_Number(self, expr, targettype=int):
        assert targettype is int
        res = z3.BitVecVal(expr.value, LONG_BIT)
        return res, True

    def visit_Attribute(self, expr, targettype=int):
        res = self._convert_attr(expr.varname, expr.attrname)
        return res, True

    def visit_MethodCall(self, expr, targettype=int):
        res, resvalid = self.visit(expr.value, IntBound)
        assert isinstance(res, IntBound)
        args = [
            self.visit(arg, arg.typ) for arg in expr.args
        ]
        methargs = [arg[0] for arg in args]
        return getattr(res, expr.methname)(*methargs), z3_and(
            resvalid, *[arg[1] for arg in args]
        )

    def visit_FuncCall(self, expr, targettype=int):
        targettypes = [int] * len(expr.args)
        args = [
            self.visit(arg, typ) for arg, typ in zip(expr.args, targettypes)
        ]
        func = globals()["z3_" + expr.funcname]
        funcargs = [arg[0] for arg in args]
        return func(*funcargs), z3_and(*[arg[1] for arg in args])

    def must_be_sat(self, rule, lhs, *conditions):
        def _find_index_to_remove():
            for removeindex in range(len(conditions)):
                if self.solver.check(z3_and(lhs == somevar, *(conditions[:removeindex] + conditions[removeindex + 1:]))) == z3.unsat:
                    return removeindex
            return -1
        todo = list(conditions)
        conditions = []
        while todo:
            c = todo.pop()
            if c is True:
                continue
            if c.decl().name() == 'and':
                todo.extend(c.children())
            else:
                conditions.append(c)

        somevar = z3.BitVec('check_not_empty', LONG_BIT)
        conditions.append(lhs == somevar)
        cond = z3_and(*conditions)
        if self.solver.check(cond) != z3.sat:
            # try to remove conditions
            while 1:
                removeindex = _find_index_to_remove()
                if removeindex >= 0:
                    del conditions[removeindex]
                    cond = z3_and(*conditions)
                else:
                    break
            raise RuleCannotApply(rule, cond, self)

    def check_rule(self, rule):
        import time
        t1 = time.time()
        print("checking %s" % rule)
        lhs, lhsvalid = self.visit(rule.pattern)
        self.must_be_sat(rule, lhs, lhsvalid)
        rhs, rhsvalid = self.visit(rule.target)
        implies_left = [lhsvalid]
        implies_right = [rhsvalid, rhs == lhs]
        for el in rule.elements:
            if isinstance(el, parse.Compute):
                expr, exprvalid = self.visit(el.expr, int)
                if el.expr.typ is not IntBound:
                    implies_left.append(self._convert_var(el.name) == expr)
                    implies_right.append(exprvalid)
                else:
                    self.name_to_intbound[el.name] = expr
                    self.glue_conditions_added.add(el.name)
                continue
            if isinstance(el, parse.Check):
                expr, _ = self.visit(el.expr, bool)
                implies_left.append(expr)
                continue
            assert 0, "unreachable"
        implies_left.extend(self.glue_conditions)
        self.must_be_sat(rule, lhs, lhsvalid, *implies_left)
        condition = z3_implies(z3_and(*implies_left), z3_and(*implies_right))
        print(condition)
        if not self.prove(condition):
            raise CouldNotProve(rule, condition, model, lhs, rhs, self)
        t2 = time.time()
        print("took %s seconds" % (t2 - t1))


def prove_source(s, force=False):
    lines = s.splitlines()
    ast = parse.parse(s)
    for rule in ast.rules:
        if rule.cantproof:
            print "SKIPPING PROOF!", rule.name
            continue
        cachename = None
        if not force:
            start_lineno = rule.sourcepos.lineno - 1
            end_lineno = rule.endsourcepos.lineno
            rule_lines = lines[start_lineno:end_lineno]

            h = md5("\n".join(rule_lines))
            cachename = os.path.join(
                CACHE_DIR, "jit_dsl_rule_%s" % (h.hexdigest(), ))
            try:
                with open(cachename, 'rb') as f:
                    f.read() # just needs to exist, really
                print "reusing previous proof", rule.name
                continue
            except IOError:
                pass
        p = Prover()
        p.check_rule(rule)
        if cachename is not None:
            try_atomic_write(cachename, "\n".join(rule_lines))
    return ast



from pypy.interpreter.gateway import unwrap_spec
from rpython.rtyper.lltypesystem import rffi
from rpython.translator.tool.cbuild import ExternalCompilationInfo
import sys

if sys.platform.startswith('darwin'):
    eci = ExternalCompilationInfo()
elif sys.platform.startswith('linux'):
    # crypt() is defined only in crypt.h on some Linux variants (eg. Fedora 28)
    eci = ExternalCompilationInfo(libraries=['crypt'], includes=["crypt.h"])
else:
    eci = ExternalCompilationInfo(libraries=['crypt'])
c_crypt = rffi.llexternal('crypt', [rffi.CCHARP, rffi.CCHARP], rffi.CCHARP,
                          compilation_info=eci, releasegil=False)

@unwrap_spec(word='text', salt='text')
def crypt(space, word, salt):
    """word will usually be a user's password. salt is a 2-character string
    which will be used to select one of 4096 variations of DES. The characters
    in salt must be either ".", "/", or an alphanumeric character. Returns
    the hashed password as a string, which will be composed of characters from
    the same alphabet as the salt."""
    res = c_crypt(word, salt)
    if not res:
        return space.w_None
    str_res = rffi.charp2str(res)
    return space.newtext(str_res)



from __future__ import absolute_import
import sys, os
from struct import pack, unpack, calcsize
from dotviewer.strunicode import tryencode, ord_byte_index

MAGIC = -0x3b83728b

CMSG_INIT        = b'i'
CMSG_START_GRAPH = b'['
CMSG_ADD_NODE    = b'n'
CMSG_ADD_EDGE    = b'e'
CMSG_ADD_LINK    = b'l'
CMSG_FIXED_FONT  = b'f'
CMSG_STOP_GRAPH  = b']'
CMSG_MISSING_LINK= b'm'
CMSG_SAY         = b's'

MSG_OK           = b'O'
MSG_ERROR        = b'E'
MSG_RELOAD       = b'R'
MSG_FOLLOW_LINK  = b'L'

# ____________________________________________________________

long_min = -2147483648
long_max = 2147483647


def message(tp, *values):
    #print >> sys.stderr, tp, values
    typecodes = [b'']
    values = list(map(tryencode, values))
    for v in values:
        if type(v) is bytes:
            typecodes.append(b'%ds' % len(v))
        elif 0 <= v < 256:
            typecodes.append(b'B')
        elif long_min <= v <= long_max:
            typecodes.append(b'l')
        else:
            typecodes.append(b'q')
    typecodes = b''.join(typecodes)
    if len(typecodes) < 256:
        return pack((b"!B%dsc" % len(typecodes)) + typecodes,
                    len(typecodes), typecodes, tp, *values)
    else:
        # too many values - encapsulate the message in another one
        return message('\x00', typecodes, pack("!c" + typecodes, tp, *values))

def decodemessage(data):
    if data:
        limit = ord_byte_index(data, 0) + 1
        if len(data) >= limit:
            typecodes = b"!c" + data[1:limit]
            end = limit + calcsize(typecodes)
            if len(data) >= end:
                msg = unpack(typecodes, data[limit:end])
                if msg[0] == b'\x00':
                    msg = unpack(b"!c" + msg[1], msg[2])
                return msg, data[end:]
            #elif end > 1000000:
            #    raise OverflowError
    return None, data

# ____________________________________________________________

class RemoteError(Exception):
    pass


class IO(object):
    _buffer = b''

    def sendmsg(self, tp, *values):
        self.sendall(message(tp, *values))

    def recvmsg(self):
        while True:
            msg, self._buffer = decodemessage(self._buffer)
            if msg is not None:
                break
            self._buffer += self.recv()
        if msg[0] != MSG_ERROR:
            return msg
        raise RemoteError(*msg[1:])


class FileIO(IO):
    def __init__(self, f_in, f_out):
        self.f_in = f_in
        self.f_out = f_out

    def sendall(self, data):
        self.f_out.write(data)
        self.f_out.flush()

    def recv(self):
        fd = self.f_in.fileno()
        data = os.read(fd, 16384)
        if not data:
            raise EOFError
        return data

    def close_sending(self):
        self.f_out.close()

    def close(self):
        self.f_out.close()
        self.f_in.close()


class SocketIO(IO):
    def __init__(self, s):
        self.s = s

    def sendall(self, data):
        self.s.sendall(data)

    def recv(self):
        data = self.s.recv(16384)
        if not data:
            raise EOFError
        return data

    def close_sending(self):
        self.s.shutdown(1)    # SHUT_WR

    def close(self):
        self.s.close()



import os
import random

import py

from pypy.interpreter.gateway import unwrap_spec, interp2app
from pypy.module.bz2.test.support import CheckAllocation


if os.name == "nt":
    from py.test import skip
    skip("bz2 module is not available on Windows")

def setup_module(mod):
    DATA = 'BZh91AY&SY.\xc8N\x18\x00\x01>_\x80\x00\x10@\x02\xff\xf0\x01\x07n\x00?\xe7\xff\xe00\x01\x99\xaa\x00\xc0\x03F\x86\x8c#&\x83F\x9a\x03\x06\xa6\xd0\xa6\x93M\x0fQ\xa7\xa8\x06\x804hh\x12$\x11\xa4i4\xf14S\xd2<Q\xb5\x0fH\xd3\xd4\xdd\xd5\x87\xbb\xf8\x94\r\x8f\xafI\x12\xe1\xc9\xf8/E\x00pu\x89\x12]\xc9\xbbDL\nQ\x0e\t1\x12\xdf\xa0\xc0\x97\xac2O9\x89\x13\x94\x0e\x1c7\x0ed\x95I\x0c\xaaJ\xa4\x18L\x10\x05#\x9c\xaf\xba\xbc/\x97\x8a#C\xc8\xe1\x8cW\xf9\xe2\xd0\xd6M\xa7\x8bXa<e\x84t\xcbL\xb3\xa7\xd9\xcd\xd1\xcb\x84.\xaf\xb3\xab\xab\xad`n}\xa0lh\tE,\x8eZ\x15\x17VH>\x88\xe5\xcd9gd6\x0b\n\xe9\x9b\xd5\x8a\x99\xf7\x08.K\x8ev\xfb\xf7xw\xbb\xdf\xa1\x92\xf1\xdd|/";\xa2\xba\x9f\xd5\xb1#A\xb6\xf6\xb3o\xc9\xc5y\\\xebO\xe7\x85\x9a\xbc\xb6f8\x952\xd5\xd7"%\x89>V,\xf7\xa6z\xe2\x9f\xa3\xdf\x11\x11"\xd6E)I\xa9\x13^\xca\xf3r\xd0\x03U\x922\xf26\xec\xb6\xed\x8b\xc3U\x13\x9d\xc5\x170\xa4\xfa^\x92\xacDF\x8a\x97\xd6\x19\xfe\xdd\xb8\xbd\x1a\x9a\x19\xa3\x80ankR\x8b\xe5\xd83]\xa9\xc6\x08\x82f\xf6\xb9"6l$\xb8j@\xc0\x8a\xb0l1..\xbak\x83ls\x15\xbc\xf4\xc1\x13\xbe\xf8E\xb8\x9d\r\xa8\x9dk\x84\xd3n\xfa\xacQ\x07\xb1%y\xaav\xb4\x08\xe0z\x1b\x16\xf5\x04\xe9\xcc\xb9\x08z\x1en7.G\xfc]\xc9\x14\xe1B@\xbb!8`'
    DATA_CRLF = 'BZh91AY&SY\xaez\xbbN\x00\x01H\xdf\x80\x00\x12@\x02\xff\xf0\x01\x07n\x00?\xe7\xff\xe0@\x01\xbc\xc6`\x86*\x8d=M\xa9\x9a\x86\xd0L@\x0fI\xa6!\xa1\x13\xc8\x88jdi\x8d@\x03@\x1a\x1a\x0c\x0c\x83 \x00\xc4h2\x19\x01\x82D\x84e\t\xe8\x99\x89\x19\x1ah\x00\r\x1a\x11\xaf\x9b\x0fG\xf5(\x1b\x1f?\t\x12\xcf\xb5\xfc\x95E\x00ps\x89\x12^\xa4\xdd\xa2&\x05(\x87\x04\x98\x89u\xe40%\xb6\x19\'\x8c\xc4\x89\xca\x07\x0e\x1b!\x91UIFU%C\x994!DI\xd2\xfa\xf0\xf1N8W\xde\x13A\xf5\x9cr%?\x9f3;I45A\xd1\x8bT\xb1<l\xba\xcb_\xc00xY\x17r\x17\x88\x08\x08@\xa0\ry@\x10\x04$)`\xf2\xce\x89z\xb0s\xec\x9b.iW\x9d\x81\xb5-+t\x9f\x1a\'\x97dB\xf5x\xb5\xbe.[.\xd7\x0e\x81\xe7\x08\x1cN`\x88\x10\xca\x87\xc3!"\x80\x92R\xa1/\xd1\xc0\xe6mf\xac\xbd\x99\xcca\xb3\x8780>\xa4\xc7\x8d\x1a\\"\xad\xa1\xabyBg\x15\xb9l\x88\x88\x91k"\x94\xa4\xd4\x89\xae*\xa6\x0b\x10\x0c\xd6\xd4m\xe86\xec\xb5j\x8a\x86j\';\xca.\x01I\xf2\xaaJ\xe8\x88\x8cU+t3\xfb\x0c\n\xa33\x13r2\r\x16\xe0\xb3(\xbf\x1d\x83r\xe7M\xf0D\x1365\xd8\x88\xd3\xa4\x92\xcb2\x06\x04\\\xc1\xb0\xea//\xbek&\xd8\xe6+t\xe5\xa1\x13\xada\x16\xder5"w]\xa2i\xb7[\x97R \xe2IT\xcd;Z\x04dk4\xad\x8a\t\xd3\x81z\x10\xf1:^`\xab\x1f\xc5\xdc\x91N\x14$+\x9e\xae\xd3\x80'

    @unwrap_spec(crlf=bool)
    def create_temp_file(space, crlf=False):
        f = py.test.ensuretemp("bz2").join("foo")
        data = (DATA, DATA_CRLF)[crlf]
        f.write(data, 'wb')

    def create_broken_temp_file():
        f = py.test.ensuretemp("bz2").join("foo")
        data = DATA[:100]
        f.write(data, 'wb')

    @unwrap_spec(data='bytes')
    def decompress(space, data):
        import popen2
        import bz2
        pop = popen2.Popen3("bunzip2", capturestderr=1)
        pop.tochild.write(data)
        pop.tochild.close()
        res = pop.fromchild.read()
        pop.fromchild.close()
        if pop.wait() != 0:
            res = bz2.decompress(data)
        return space.wrap(res)

    mod.TEXT = 'root:x:0:0:root:/root:/bin/bash\nbin:x:1:1:bin:/bin:\ndaemon:x:2:2:daemon:/sbin:\nadm:x:3:4:adm:/var/adm:\nlp:x:4:7:lp:/var/spool/lpd:\nsync:x:5:0:sync:/sbin:/bin/sync\nshutdown:x:6:0:shutdown:/sbin:/sbin/shutdown\nhalt:x:7:0:halt:/sbin:/sbin/halt\nmail:x:8:12:mail:/var/spool/mail:\nnews:x:9:13:news:/var/spool/news:\nuucp:x:10:14:uucp:/var/spool/uucp:\noperator:x:11:0:operator:/root:\ngames:x:12:100:games:/usr/games:\ngopher:x:13:30:gopher:/usr/lib/gopher-data:\nftp:x:14:50:FTP User:/var/ftp:/bin/bash\nnobody:x:65534:65534:Nobody:/home:\npostfix:x:100:101:postfix:/var/spool/postfix:\nniemeyer:x:500:500::/home/niemeyer:/bin/bash\npostgres:x:101:102:PostgreSQL Server:/var/lib/pgsql:/bin/bash\nmysql:x:102:103:MySQL server:/var/lib/mysql:/bin/bash\nwww:x:103:104::/var/www:/bin/false\n'
    mod.DATA = DATA
    mod.DATA_CRLF = DATA_CRLF
    mod.create_temp_file = create_temp_file
    mod.decompress = decompress
    mod.create_broken_temp_file = create_broken_temp_file
    s = 'abcdefghijklmnop'
    mod.RANDOM_DATA = ''.join([s[int(random.random() * len(s))] for i in range(30000)])


class AppTestBZ2File(CheckAllocation):
    spaceconfig = {
        'usemodules': ['bz2', 'binascii', 'time', 'struct']
    }

    def setup_class(cls):
        cls.w_TEXT = cls.space.newbytes(TEXT)
        cls.w_DATA = cls.space.newbytes(DATA)
        cls.w_DATA_CRLF = cls.space.newbytes(DATA_CRLF)
        cls.w_temppath = cls.space.wrap(
            str(py.test.ensuretemp("bz2").join("foo")))
        if cls.runappdirect:
            cls.w_create_temp_file = create_temp_file
            cls.w_create_broken_temp_file = lambda self: create_broken_temp_file()
            cls.w_decompress = lambda self, *args: decompress(cls.space, *args)
        else:
            cls.w_create_temp_file = cls.space.wrap(interp2app(create_temp_file))
            cls.w_create_broken_temp_file = cls.space.wrap(interp2app(create_broken_temp_file))
            cls.w_decompress = cls.space.wrap(interp2app(decompress))
        cls.w_random_data = cls.space.wrap(RANDOM_DATA)

    def test_attributes(self):
        from bz2 import BZ2File

        bz2f = BZ2File(self.temppath, mode="w")
        assert bz2f.name == self.temppath
        assert bz2f.newlines == None
        assert bz2f.mode == "wb"
        assert bz2f.softspace == False
        assert bz2f.closed == False
        bz2f.close()
        assert bz2f.closed == True

    def test_creation(self):
        from bz2 import BZ2File

        raises(ValueError, BZ2File, self.temppath, mode='w', compresslevel=10)
        raises(ValueError, BZ2File, self.temppath, mode='XYZ')
        # XXX the following is fine, currently:
        #raises(ValueError, BZ2File, self.temppath, mode='ww')

        BZ2File(self.temppath, mode='wU', buffering=0, compresslevel=8)
        BZ2File(self.temppath, mode='wb')
        # a large buf size
        BZ2File(self.temppath, mode='w', buffering=4096)

        exc = raises(IOError, BZ2File, 'xxx', 'r')
        assert "'xxx'" in str(exc.value)

    def test_close(self):
        from bz2 import BZ2File

        # writeonly
        bz2f = BZ2File(self.temppath, mode='w')
        bz2f.close()
        bz2f.close()

        # readonly
        bz2f = BZ2File(self.temppath, mode='r')
        bz2f.close()
        bz2f.close()

    def test_tell(self):
        from bz2 import BZ2File

        bz2f = BZ2File(self.temppath, mode='w')
        bz2f.close()
        raises(ValueError, bz2f.tell)

        bz2f = BZ2File(self.temppath, mode='w')
        pos = bz2f.tell()
        bz2f.close()
        assert pos == 0

    def test_seek(self):
        from bz2 import BZ2File

        # hack to create a foo file
        open(self.temppath, "w").close()

        # cannot seek if close
        bz2f = BZ2File(self.temppath, mode='r')
        bz2f.close()
        raises(ValueError, bz2f.seek, 0)

        # cannot seek if 'w'
        bz2f = BZ2File(self.temppath, mode='w')
        raises(IOError, bz2f.seek, 0)
        bz2f.close()

        bz2f = BZ2File(self.temppath, mode='r')
        raises(TypeError, bz2f.seek)
        raises(TypeError, bz2f.seek, "foo")
        raises(TypeError, bz2f.seek, 0, "foo")

        bz2f.seek(0)
        assert bz2f.tell() == 0
        bz2f.close()
        del bz2f   # delete from this frame, which is captured in the traceback

    def test_open_close_del(self):
        from bz2 import BZ2File
        self.create_temp_file()

        for i in range(10):
            f = BZ2File(self.temppath)
            f.close()
            del f

    def test_open_non_existent(self):
        from bz2 import BZ2File
        raises(IOError, BZ2File, "/non/existent/path")

    def test_open_mode_U(self):
        # bug #1194181: bz2.BZ2File opened for write with mode "U"
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath, "U")
        bz2f.close()
        f = open(self.temppath)
        f.seek(0, 2)
        f.read()
        assert f.tell() == len(self.DATA)
        f.close()

    def test_seek_forward(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        bz2f.seek(150) # (150, 0)
        assert bz2f.read() == self.TEXT[150:]
        bz2f.close()

    def test_seek_backwards(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        bz2f.read(500)
        bz2f.seek(-150, 1)
        assert bz2f.read() == self.TEXT[500 - 150:]
        bz2f.close()

    def test_seek_backwards_from_end(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        bz2f.seek(-150, 2)
        assert bz2f.read() == self.TEXT[len(self.TEXT) - 150:]
        bz2f.close()

    def test_seek_post_end(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        bz2f.seek(150000)
        assert bz2f.tell() == len(self.TEXT)
        assert bz2f.read() == b""
        bz2f.close()

    def test_seek_post_end_twice(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        bz2f.seek(150000)
        bz2f.seek(150000)
        assert bz2f.tell() == len(self.TEXT)
        assert bz2f.read() == b""
        bz2f.close()

    def test_seek_pre_start(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        bz2f.seek(-150)
        assert bz2f.tell() == 0
        assert bz2f.read() == self.TEXT
        bz2f.close()

    def test_readline(self):
        from bz2 import BZ2File
        from cStringIO import StringIO
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        raises(TypeError, bz2f.readline, None)
        sio = StringIO(self.TEXT)
        for line in sio.readlines():
            line_read = bz2f.readline()
            assert line_read == line
        bz2f.close()

    def test_read(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        raises(TypeError, bz2f.read, None)
        text_read = bz2f.read()
        assert text_read == self.TEXT
        bz2f.close()

    def test_silently_closes(self):
        from bz2 import BZ2File
        self.create_broken_temp_file()
        BZ2File(self.temppath)
        # check that no C-level malloc is left behind

    def test_read_broken_file(self):
        from bz2 import BZ2File
        self.create_broken_temp_file()
        bz2f = BZ2File(self.temppath)
        raises(EOFError, bz2f.read)
        bz2f.close()

    def test_subsequent_read_broken_file(self):
        from bz2 import BZ2File
        counter = 0
        self.create_broken_temp_file()
        bz2f = BZ2File(self.temppath)
        try:
            bz2f.read(10)
            counter += 1
            if counter > 100:
                raise Exception("should generate EOFError earlier")
        except EOFError:
            pass
        bz2f.close()

    def test_read_chunk9(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        text_read = b""
        while True:
            data = bz2f.read(9) # 9 doesn't divide evenly into data length
            if not data:
                break
            text_read += data
        assert text_read == self.TEXT
        bz2f.close()

    def test_read_100_bytes(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        assert bz2f.read(100) == self.TEXT[:100]
        bz2f.close()

    def test_universal_newlines_lf(self):
        from bz2 import BZ2File
        self.create_temp_file()

        bz2f = BZ2File(self.temppath, "rU")
        assert bz2f.read() == self.TEXT
        assert bz2f.newlines == "\n"
        bz2f.close()

    def test_universal_newlines_crlf(self):
        from bz2 import BZ2File
        self.create_temp_file(crlf=True)

        bz2f = BZ2File(self.temppath, "rU")
        data = bz2f.read()
        assert data == self.TEXT
        assert bz2f.newlines == "\r\n"
        bz2f.close()

    def test_readlines(self):
        from bz2 import BZ2File
        from cStringIO import StringIO
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        raises(TypeError, bz2f.readlines, None)
        sio = StringIO(self.TEXT)
        assert bz2f.readlines() == sio.readlines()
        bz2f.close()

    def test_iterator(self):
        from bz2 import BZ2File
        from cStringIO import StringIO
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        sio = StringIO(self.TEXT)
        assert list(iter(bz2f)) == sio.readlines()
        bz2f.close()

    def test_xreadlines(self):
        from bz2 import BZ2File
        from cStringIO import StringIO
        self.create_temp_file()

        bz2f = BZ2File(self.temppath)
        sio = StringIO(self.TEXT)
        assert list(bz2f.xreadlines()) == sio.readlines()
        bz2f.close()

    def test_readlines_bug_1191043(self):
        # readlines()/xreadlines() for files containing no newline
        from bz2 import BZ2File

        DATA = 'BZh91AY&SY\xd9b\x89]\x00\x00\x00\x03\x80\x04\x00\x02\x00\x0c\x00 \x00!\x9ah3M\x13<]\xc9\x14\xe1BCe\x8a%t'
        f = open(self.temppath, "wb")
        f.write(DATA)
        f.close()

        bz2f = BZ2File(self.temppath)
        lines = bz2f.readlines()
        bz2f.close()
        assert lines == ['Test']

        bz2f = BZ2File(self.temppath)
        xlines = list(bz2f.xreadlines())
        bz2f.close()
        assert xlines == ['Test']

    def test_write(self):
        from bz2 import BZ2File

        bz2f = BZ2File(self.temppath, 'w')
        raises(TypeError, bz2f.write)
        bz2f.write(self.TEXT)
        bz2f.close()

        f = open(self.temppath, "rb")
        assert self.decompress(f.read()) == self.TEXT
        f.close()

    def test_write_chunks_10(self):
        from bz2 import BZ2File

        bz2f = BZ2File(self.temppath, 'w')
        n = 0
        while True:
            data = self.TEXT[n * 10:(n + 1) * 10]
            if not data:
                break

            bz2f.write(data)
            n += 1
        bz2f.close()

        f = open(self.temppath, "rb")
        assert self.decompress(f.read()) == self.TEXT
        f.close()

    def test_writelines(self):
        from bz2 import BZ2File
        from cStringIO import StringIO

        bz2f = BZ2File(self.temppath, 'w')
        raises(TypeError, bz2f.writelines)
        sio = StringIO(self.TEXT)
        bz2f.writelines(sio.readlines())
        bz2f.close()
        f = open(self.temppath, "rb")
        assert self.decompress(f.read()) == self.TEXT
        f.close()

    def test_write_methods_on_readonly_file(self):
        from bz2 import BZ2File

        bz2f = BZ2File(self.temppath, 'r')
        raises(IOError, bz2f.write, b"abc")
        raises(IOError, bz2f.writelines, [b"abc"])
        bz2f.close()

    def test_write_bigger_file(self):
        from bz2 import BZ2File
        import random
        bz2f = BZ2File(self.temppath, 'w')
        bz2f.write(self.random_data)
        bz2f.close()
        bz2f = BZ2File(self.temppath, 'r')
        assert bz2f.read() == self.random_data
        del bz2f   # delete from this frame, which is captured in the traceback

    def test_context_manager(self):
        from bz2 import BZ2File

        with BZ2File(self.temppath, 'w') as f:
            assert not f.closed
            f.write(b"abc")
        assert f.closed
        with BZ2File(self.temppath, 'r') as f:
            data = f.read()
            assert data == b"abc"
        assert f.closed



# has_cmdline_bunzip2 = sys.platform not in ("win32", "os2emx", "riscos")
# 
# if has_cmdline_bunzip2:
#     def decompress(self, data):
#         pop = popen2.Popen3("bunzip2", capturestderr=1)
#         pop.tochild.write(data)
#         pop.tochild.close()
#         ret = pop.fromchild.read()
#         pop.fromchild.close()
#         if pop.wait() != 0:
#             ret = bz2.decompress(data)
#         return ret
# 
# else:
#     # popen2.Popen3 doesn't exist on Windows, and even if it did, bunzip2
#     # isn't available to run.
#     def decompress(self, data):
#         return bz2.decompress(data)



from rpython.tool.sourcetools import compile2
from rpython.rlib.rarithmetic import r_uint
from rpython.jit.backend.ppc.form import IDesc, IDupDesc
from rpython.jit.backend.ppc.ppc_field import IField

##     "opcode": ( 0,  5),
##     "rA":     (11, 15, 'unsigned', regname._R),
##     "rB":     (16, 20, 'unsigned', regname._R),
##     "Rc":     (31, 31),
##     "rD":     ( 6, 10, 'unsigned', regname._R),
##     "OE":     (21, 21),
##     "XO2":    (22, 30),

## XO = Form("rD", "rA", "rB", "OE", "XO2", "Rc")

##     add   = XO(31, XO2=266, OE=0, Rc=0)

##     def add(rD, rA, rB):
##         v = 0
##         v |= (31&(2**(5-0+1)-1)) << (32-5-1)
##         ...
##         return v

def make_func(name, desc):
    sig = []
    fieldvalues = []
    for field in desc.fields:
        if field in desc.specializations:
            fieldvalues.append((field, desc.specializations[field]))
        else:
            sig.append(field.name)
            fieldvalues.append((field, field.name))
    if isinstance(desc, IDupDesc):
        for destfield, srcfield in desc.dupfields.iteritems():
            fieldvalues.append((destfield, srcfield.name))
    body = ['v = r_uint(0)']
    assert 'v' not in sig # that wouldn't be funny
    #body.append('print %r'%name + ', ' + ', '.join(["'%s:', %s"%(s, s) for s in sig]))
    for field, value in fieldvalues:
        if field.name == 'spr':
            body.append('spr1 = (%s&31) << 5 | (%s >> 5 & 31)'%(value, value))
            value = 'spr1'
        elif field.name == 'mbe':
            body.append('mbe1 = (%s & 31) << 1 | (%s & 32) >> 5' % (value, value))
            value = 'mbe1'
        elif field.name == 'sh':
            body.append('sh1 = (%s & 31) << 10 | (%s & 32) >> 5' % (value, value))
            value = 'sh1'
        elif field.name == 'fvrT':
            body.append('vrT1 = (%s & 31) << 21 | (%s & 32) >> 5' % (value, value))
            value = 'vrT1'
        elif field.name == 'fvrA':
            body.append('fvrA1 = (%s & 31) << 14 | (%s & 32) >> 5' % (value, value))
            value = 'fvrA1'
        elif field.name == 'fvrB':
            body.append('fvrB1 = (%s & 31) << 10 | (%s & 32) >> 5' % (value, value))
            value = 'fvrB1'
        if isinstance(field, IField):
            body.append('v |= ((%3s >> 2) & r_uint(%#05x)) << 2' % (value, field.mask))
        else:
            body.append('v |= (%3s & r_uint(%#05x)) << %d'%(value,
                                                            field.mask,
                                                            (32 - field.right - 1)))
    #body.append('self.check(desc, v, %s)' % ', '.join(sig))
    body.append('self.emit(v)')
    src = 'def %s(self, %s):\n    %s'%(name, ', '.join(sig), '\n    '.join(body))
    d = {'r_uint':r_uint, 'desc': desc}
    #print src
    exec(compile2(src), d)
    return d[name]

def make_rassembler(cls):
    # XXX tooons of very-old code patched to get medium-old code patched
    # to get newer code :-(
    bases = [make_rassembler(b) for b in cls.__bases__]
    ns = {}
    for k, v in cls.__dict__.iteritems():
        if isinstance(v, IDesc):
            v = make_func(k, v)
        ns[k] = v
    rcls = type('R' + cls.__name__, tuple(bases), ns)
    def emit(self, value):
        self.write32(value)
    rcls.emit = emit
    return rcls



import sys
from rpython.rlib.objectmodel import we_are_translated, fetch_translated_config
from rpython.rlib.objectmodel import specialize
from rpython.rlib.rarithmetic import r_longlong
from rpython.rtyper.lltypesystem import lltype, llmemory, rstr
from rpython.rtyper.lltypesystem.lloperation import llop
from rpython.rtyper.extregistry import ExtRegistryEntry
from rpython.rtyper.annlowlevel import llhelper, hlstr
from rpython.rtyper.annlowlevel import cast_gcref_to_instance
from rpython.rtyper.lltypesystem import lltype, rffi


CMD_PRINT       = 1
CMD_BACKTRACE   = 2
CMD_LOCALS      = 3
CMD_BREAKPOINTS = 4
CMD_STACKID     = 5
CMD_ATTACHID    = 6
CMD_COMPILEWATCH= 7
CMD_CHECKWATCH  = 8
CMD_WATCHVALUES = 9
ANSWER_LINECACHE= 19
ANSWER_TEXT     = 20
ANSWER_STACKID  = 21
ANSWER_NEXTNID  = 22
ANSWER_WATCH    = 23
ANSWER_CHBKPT   = 24


def stop_point(place=0):
    """Indicates a point in the execution of the RPython program where
    the reverse-debugger can stop.  When reverse-debugging, we see
    the "time" as the index of the stop-point that happened.
    """
    if we_are_translated():
        if fetch_translated_config().translation.reverse_debugger:
            llop.revdb_stop_point(lltype.Void, place)

def register_debug_command(command, lambda_func):
    """Register the extra RPython-implemented debug command."""

def send_answer(cmd, arg1=0, arg2=0, arg3=0, extra=""):
    """For RPython debug commands: writes an answer block to stdout"""
    llop.revdb_send_answer(lltype.Void, cmd, arg1, arg2, arg3, extra)

def send_output(text):
    send_answer(ANSWER_TEXT, extra=text)

def send_print(text):
    send_answer(ANSWER_TEXT, 1, extra=text)   # adds a newline

def send_nextnid(unique_id):
    send_answer(ANSWER_NEXTNID, unique_id)

def send_watch(text, ok_flag):
    send_answer(ANSWER_WATCH, ok_flag, extra=text)

def send_linecache(filename, linenum, strip=True):
    send_answer(ANSWER_LINECACHE, linenum, int(strip), extra=filename)

def send_change_breakpoint(breakpointnum, newtext=''):
    send_answer(ANSWER_CHBKPT, breakpointnum, extra=newtext)

def current_time():
    """For RPython debug commands: returns the current time."""
    return llop.revdb_get_value(lltype.SignedLongLong, 'c')

def current_break_time():
    """Returns the time configured for the next break.  When going forward,
    this is the target time at which we'll stop going forward."""
    return llop.revdb_get_value(lltype.SignedLongLong, 'b')

def total_time():
    """For RPython debug commands: returns the total time (measured
    as the total number of stop-points)."""
    return llop.revdb_get_value(lltype.SignedLongLong, 't')

def currently_created_objects():
    """For RPython debug commands: returns the current value of
    the object creation counter.  All objects created so far have
    a lower unique id; all objects created afterwards will have a
    unique id greater or equal."""
    return llop.revdb_get_value(lltype.SignedLongLong, 'u')

def current_place():
    """For RPython debug commands: the value of the 'place' argument
    passed to stop_point().
    """
    return llop.revdb_get_value(lltype.Signed, 'p')

def flag_io_disabled():
    """Returns True if we're in the debugger typing commands."""
    if we_are_translated():
        if fetch_translated_config().translation.reverse_debugger:
            flag = llop.revdb_get_value(lltype.Signed, 'i')
            return flag != ord('R')  # FID_REGULAR_MODE
    return False

## @specialize.arg(1)
## def go_forward(time_delta, callback):
##     """For RPython debug commands: tells that after this function finishes,
##     the debugger should run the 'forward <time_delta>' command and then
##     invoke the 'callback' with no argument.
##     """
##     _change_time('f', time_delta, callback)

def breakpoint(num):
    llop.revdb_breakpoint(lltype.Void, num)

def set_thread_breakpoint(tnum):
    llop.revdb_set_thread_breakpoint(lltype.Void, tnum)

@specialize.argtype(0)
def get_unique_id(x):
    """Returns the creation number of the object 'x'.  For objects created
    by the program, it is globally unique, monotonic, and reproducible
    among multiple processes.  For objects created by a debug command,
    this returns a (random) negative number.  Right now, this returns 0
    for all prebuilt objects.
    """
    return llop.revdb_get_unique_id(lltype.SignedLongLong, x)

def track_object(unique_id, callback):
    """Track the creation of the object given by its unique_id, which must
    be in the future (i.e. >= currently_created_objects()).  Call this
    before go_forward().  If go_forward() goes over the creation of this
    object, then 'callback(gcref)' is called.  Careful in callback(),
    gcref is not fully initialized and should not be immediately read from,
    only stored for later.  The purpose of callback() is to possibly
    call track_object() again to track the next object, and/or to call
    breakpoint().  Note: object tracking remains activated until one of:
    (1) we reach the creation time in go_forward(); (2) we call
    track_object() to track a different object; (3) we call jump_in_time().
    """
    ll_callback = llhelper(_CALLBACK_GCREF_FNPTR, callback)
    llop.revdb_track_object(lltype.Void, unique_id, ll_callback)

def watch_save_state(force=False):
    return llop.revdb_watch_save_state(lltype.Bool, force)

def watch_restore_state(any_watch_point):
    llop.revdb_watch_restore_state(lltype.Void, any_watch_point)


def split_breakpoints_arg(breakpoints):
    # RPython generator to help in splitting the string arg in CMD_BREAKPOINTS
    n = 0
    i = 0
    while i < len(breakpoints):
        kind = breakpoints[i]
        i += 1
        if kind != '\x00':
            length = (ord(breakpoints[i]) |
                      (ord(breakpoints[i + 1]) << 8) |
                      (ord(breakpoints[i + 2]) << 16))
            assert length >= 0
            i += 3
            yield n, kind, breakpoints[i : i + length]
            i += length
        n += 1
    assert i == len(breakpoints)


# ____________________________________________________________


## @specialize.arg(2)
## def _change_time(mode, time, callback):
##     ll_callback = llhelper(_CALLBACK_NOARG_FNPTR, callback)
##     llop.revdb_change_time(lltype.Void, mode, time, ll_callback)
## _CALLBACK_NOARG_FNPTR = lltype.Ptr(lltype.FuncType([], lltype.Void))
_CALLBACK_GCREF_FNPTR = lltype.Ptr(lltype.FuncType([llmemory.GCREF],
                                                   lltype.Void))
_CMDPTR = rffi.CStructPtr('rpy_revdb_command_s',
                          ('cmd', rffi.INT),
                          ('arg1', lltype.SignedLongLong),
                          ('arg2', lltype.SignedLongLong),
                          ('arg3', lltype.SignedLongLong),
                          hints={'ignore_revdb': True})


class RegisterDebugCommand(ExtRegistryEntry):
    _about_ = register_debug_command

    def compute_result_annotation(self, s_command_num, s_lambda_func):
        command_num = s_command_num.const
        lambda_func = s_lambda_func.const
        assert isinstance(command_num, (int, str))
        t = self.bookkeeper.annotator.translator
        if t.config.translation.reverse_debugger:
            func = lambda_func()
            try:
                cmds = t.revdb_commands
            except AttributeError:
                cmds = t.revdb_commands = {}
            old_func = cmds.setdefault(command_num, func)
            assert old_func is func
            s_func = self.bookkeeper.immutablevalue(func)
            arg_getter = getattr(self, 'arguments_' + str(command_num),
                                 self.default_arguments)
            self.bookkeeper.emulate_pbc_call(self.bookkeeper.position_key,
                                             s_func, arg_getter())

    def default_arguments(self):
        from rpython.annotator import model as annmodel
        from rpython.rtyper import llannotation
        return [llannotation.SomePtr(ll_ptrtype=_CMDPTR),
                annmodel.SomeString()]

    def arguments_ALLOCATING(self):
        from rpython.annotator import model as annmodel
        from rpython.rtyper import llannotation
        return [annmodel.SomeInteger(knowntype=r_longlong),
                llannotation.lltype_to_annotation(llmemory.GCREF)]

    def arguments_WATCHING(self):
        raise Exception("XXX remove me")

    def specialize_call(self, hop):
        hop.exception_cannot_occur()


# ____________________________________________________________

# Emulation for strtod() and dtoa() when running debugger commands
# (we can't easily just call C code there).  The emulation can return
# a crude result.  Hack hack hack.

_INVALID_STRTOD = -3.46739514239368e+113

def emulate_strtod(input):
    d = llop.revdb_strtod(lltype.Float, input)
    if d == _INVALID_STRTOD:
        raise ValueError
    return d

def emulate_dtoa(value):
    s = llop.revdb_dtoa(lltype.Ptr(rstr.STR), value)
    s = hlstr(s)
    assert s is not None
    return s

def emulate_modf(x):
    return (llop.revdb_modf(lltype.Float, x, 0),
            llop.revdb_modf(lltype.Float, x, 1))

def emulate_frexp(x):
    return (llop.revdb_frexp(lltype.Float, x, 0),
            int(llop.revdb_frexp(lltype.Float, x, 1)))



import py, os, sys
from pypy.module._cppyy import interp_cppyy, executor


class AppTestREGRESSION:
    spaceconfig = dict(usemodules=['_cppyy', '_rawffi', 'itertools'])

    def setup_class(cls):
        cls.w_example01 = cls.space.appexec([], """():
            import ctypes, _cppyy
            _cppyy._post_import_startup()""")

    def test01_dir(self):
        """These dir() methods used to crash."""

        import _cppyy as cppyy

        cppyy.gbl.gInterpreter.Declare("namespace cppyy_regression_test { void iii() {}; }")

        assert not 'iii' in cppyy.gbl.cppyy_regression_test.__dict__
        assert not '__abstractmethods__' in dir(cppyy.gbl.cppyy_regression_test)
        assert '__class__' in dir(cppyy.gbl.cppyy_regression_test)
        assert 'iii' in dir(cppyy.gbl.cppyy_regression_test)

        assert not 'iii' in cppyy.gbl.cppyy_regression_test.__dict__
        assert cppyy.gbl.cppyy_regression_test.iii
        assert 'iii' in cppyy.gbl.cppyy_regression_test.__dict__

    def test02_default_template_arguments(self):
        """Calling a templated method on a templated class with all defaults used to crash."""

        import _cppyy as cppyy

        cppyy.gbl.gInterpreter.Declare("""
        template<typename T>
        class AllDefault {
        public:
           AllDefault(int val) : m_t(val) {}
           template<int aap=1, int noot=2>
              int do_stuff() { return m_t+aap+noot; }

        public:
           T m_t;
        };""")

        a = cppyy.gbl.AllDefault[int](24)
        a.m_t = 21;
        assert a.do_stuff() == 24
